from model_bakery import baker
from datetime import timedelta
from django.test import TestCase
from mkite.orm.base.models import (
    ChemNode,
    CalcNode,
    Formula,
    Elements,
)


class TestBase(TestCase):
    def test_chem(self):
        node = baker.make(ChemNode)
        self.assertTrue(hasattr(node, "parentjob"))

    def test_calc(self):
        node = baker.make(CalcNode)
        self.assertTrue(hasattr(node, "parentjob"))
        self.assertTrue(hasattr(node, "chemnode"))
        self.assertTrue(isinstance(node.chemnode, ChemNode))

    def test_formula(self):
        formula = baker.make(Formula)
        self.assertTrue(hasattr(formula, "name"))



from typing import Iterable, List
from abc import ABC, abstractmethod

from mkite_core.models import NodeResults, JobResults


class DbImporterError(Exception):
    pass


class DbImporter(ABC):
    """Class that provides translation between mkite and other databases."""

    RECIPE_DICT = None
    PACKAGE_DICT = None

    def __init__(
        self,
        project: str,
        experiment: str,
    ):
        self.project = project
        self.experiment = experiment

    @classmethod
    def from_env(cls, project: str, experiment: str) -> "DbImporter":
        """Creates a new instance of DbImporter based on environmental
        variables. This is useful when parsing API keys and similar
        information that does not have to be handled by the command.
        By default, no environmental variables are assumed and the
        DbImporter is instantiated as normal"""
        return cls(project, experiment)

    @abstractmethod
    def query(self, *args, **kwargs) -> Iterable[dict]:
        """Queries the database and returns a list of results"""

    @abstractmethod
    def convert(self, parsed: List[dict]) -> List[NodeResults]:
        """Converts the parsed results from the database into the
        mkite standard, using the NodeResults notation"""

    def query_and_convert(self, *args, **kwargs) -> List[NodeResults]:
        raw_data = self.query(*args, **kwargs)
        nodes = self.convert(raw_data)
        return nodes

    def import_data(self, **kwargs) -> JobResults:
        # TODO: add RunStats for querying if possible
        job_data = self.create_job(
            self.project,
            self.experiment,
            isroot=True,
            options=kwargs,
        )
        nodes = self.query_and_convert(**kwargs)

        return JobResults(job=job_data, nodes=nodes)

    def create_job(
        self,
        project: str,
        experiment: str,
        isroot=True,
        options={},
    ):
        """Creates a job that will be used to deserialize all the jobs
        added to the database through the given options"""

        recipe_dict = {
            "package": self.PACKAGE_DICT,
            **self.RECIPE_DICT,
        }

        jobdict = {
            "experiment": {
                "name": experiment,
                "project": {"name": project},
            },
            "recipe": recipe_dict,
            "isroot": isroot,
            "status": "D",
            "options": options,
        }

        return jobdict



import uuid
import warnings
from itertools import chain

from django.db import models, transaction
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType


class DbEntry(models.Model):
    """Abstract base class for every unique, identifiable entry in the database"""

    uuid = models.UUIDField(unique=True, default=uuid.uuid4)
    ctime = models.DateTimeField(db_index=True, auto_now_add=True)
    mtime = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Node(DbEntry):
    class Meta:
        abstract = True

    def as_dict(self):
        opts = self._meta
        data = {"@module": self.__class__.__module__, "@class": self.__class__.__name__}

        for f in chain(opts.concrete_fields, opts.private_fields):
            data[f.name] = f.value_from_object(self)

        for f in opts.many_to_many:
            data[f.name] = [i.id for i in f.value_from_object(self)]

        if hasattr(self, "uuid"):
            data["uuid"] = str(self.uuid)

        if "tagged_items" in data:
            data.pop("tagged_items")

        return data

    def get_data(self):
        """Nodes are generic entities that can be subclassed. When
        getting their data, however, we are interested in serializing
        all the data from the node, including the models related
        by OneToOne fields. This method verifies what are the non-null
        OneToOne fields containing data and returns the data associated
        to this node.
        """
        fields = [f for f in self._meta.related_objects if f.one_to_one]

        for f in fields:
            if hasattr(self, f.name):
                return getattr(self, f.name).as_dict()

        return self.as_dict()


class ChemNode(Node):
    """Base class for every chemical information point in the database.

    Nodes can only be created by Jobs. To ensure the traceability
    of chemical data, every Node has to have a `parentjob`.

    Nodes are abstract classes that have a `data` generic foreign key. This tries to connect "unstructured" data
    to the node itself at the expense of more storage space. As the multi-table inheritance remains
    in place, the generic foreign key contains redundant information. However, it facilitates finding out what is the
    information that is stored in the node.
    """

    parentjob = models.ForeignKey(
        "jobs.Job",
        null=False,
        related_name="chemnodes",
        on_delete=models.CASCADE,
    )

    def as_dict(self):
        data = super().as_dict()
        data["uuid"] = str(data["uuid"])
        return data


class CalcNode(Node):
    """Base class for every calculation in the database. This includes
    energies, forces, analysis, descriptors etc.

    Nodes can only be created by Jobs. To ensure the traceability
    of data nodes, every Node has to have a `parentjob`. Furthermore,
    every CalcNode has to point to a single ChemNode, to which the information
    is associated. Otherwise, it is just floating information in the database.
    """

    parentjob = models.ForeignKey(
        "jobs.Job",
        null=False,
        related_name="calcnodes",
        on_delete=models.CASCADE,
    )

    chemnode = models.ForeignKey(
        ChemNode,
        null=False,
        related_name="calcnodes",
        on_delete=models.CASCADE,
    )

    def as_dict(self):
        data = super().as_dict()
        data["uuid"] = str(data["uuid"])
        return data


class Formula(models.Model):
    name = models.CharField(max_length=128, unique=True)
    charge = models.SmallIntegerField(default=0)

    def as_info(self):
        from mkite_core.models import FormulaInfo

        return FormulaInfo(self.name, self.charge)


class Elements(models.TextChoices):
    H = "H"
    He = "He"
    Li = "Li"
    Be = "Be"
    B = "B"
    C = "C"
    N = "N"
    O = "O"
    F = "F"
    Ne = "Ne"
    Na = "Na"
    Mg = "Mg"
    Al = "Al"
    Si = "Si"
    P = "P"
    S = "S"
    Cl = "Cl"
    Ar = "Ar"
    K = "K"
    Ca = "Ca"
    Sc = "Sc"
    Ti = "Ti"
    V = "V"
    Cr = "Cr"
    Mn = "Mn"
    Fe = "Fe"
    Co = "Co"
    Ni = "Ni"
    Cu = "Cu"
    Zn = "Zn"
    Ga = "Ga"
    Ge = "Ge"
    As = "As"
    Se = "Se"
    Br = "Br"
    Kr = "Kr"
    Rb = "Rb"
    Sr = "Sr"
    Y = "Y"
    Zr = "Zr"
    Nb = "Nb"
    Mo = "Mo"
    Tc = "Tc"
    Ru = "Ru"
    Rh = "Rh"
    Pd = "Pd"
    Ag = "Ag"
    Cd = "Cd"
    In = "In"
    Sn = "Sn"
    Sb = "Sb"
    Te = "Te"
    I = "I"
    Xe = "Xe"
    Cs = "Cs"
    Ba = "Ba"
    La = "La"
    Ce = "Ce"
    Pr = "Pr"
    Nd = "Nd"
    Pm = "Pm"
    Sm = "Sm"
    Eu = "Eu"
    Gd = "Gd"
    Tb = "Tb"
    Dy = "Dy"
    Ho = "Ho"
    Er = "Er"
    Tm = "Tm"
    Yb = "Yb"
    Lu = "Lu"
    Hf = "Hf"
    Ta = "Ta"
    W = "W"
    Re = "Re"
    Os = "Os"
    Ir = "Ir"
    Pt = "Pt"
    Au = "Au"
    Hg = "Hg"
    Tl = "Tl"
    Pb = "Pb"
    Bi = "Bi"
    Po = "Po"
    At = "At"
    Rn = "Rn"
    Fr = "Fr"
    Ra = "Ra"
    Ac = "Ac"
    Th = "Th"
    Pa = "Pa"
    U = "U"
    Np = "Np"
    Pu = "Pu"
    Am = "Am"
    Cm = "Cm"
    Bk = "Bk"
    Cf = "Cf"
    Es = "Es"
    Fm = "Fm"
    Md = "Md"
    No = "No"
    Lr = "Lr"
    Rf = "Rf"
    Db = "Db"
    Sg = "Sg"
    Bh = "Bh"
    Hs = "Hs"
    Mt = "Mt"
    Ds = "Ds"
    Rg = "Rg"
    Cn = "Cn"
    Nh = "Nh"
    Fl = "Fl"
    Mc = "Mc"
    Lv = "Lv"
    Ts = "Ts"
    Og = "Og"



from django.db import models
from django.contrib.postgres.fields import ArrayField

from taggit.managers import TaggableManager
from mkite.orm.repr import _named_repr
from mkite.orm.base.models import DbEntry, ChemNode, Elements, Formula


class Molecule(ChemNode):
    """Class to hold the information of molecular graphs"""

    formula = models.ForeignKey(
        Formula,
        on_delete=models.PROTECT,
        related_name="molecules",
    )

    inchikey = models.CharField(
        max_length=27,
        null=False,
        unique=True,
    )

    smiles = models.CharField(
        max_length=10000,
        null=False,
        unique=True,
    )

    siteprops = models.JSONField(default=dict)

    attributes = models.JSONField(default=dict)

    tags = TaggableManager()

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.inchikey} ({self.id})>"

    def as_info(self):
        from mkite_core.models import MoleculeInfo

        return MoleculeInfo.from_molecule(self)


class Conformer(ChemNode):
    formula = models.ForeignKey(
        Formula,
        on_delete=models.PROTECT,
        related_name="conformers",
    )

    mol = models.ForeignKey(
        Molecule,
        null=True,
        related_name="conformers",
        on_delete=models.PROTECT,
    )

    species = ArrayField(
        models.CharField(
            max_length=2,
            null=False,
            choices=Elements.choices,
        )
    )
    coords = ArrayField(ArrayField(models.FloatField(), size=3))
    siteprops = models.JSONField(default=dict)
    attributes = models.JSONField(default=dict)

    def __repr__(self):
        ikey = "None" if self.mol is None else self.mol.inchikey
        return (
            f"<{self.__class__.__name__}: {self.formula.name}, Mol {ikey} ({self.id})>"
        )

    def as_info(self):
        from mkite_core.models import ConformerInfo

        return ConformerInfo.from_conformer(self)



import json
import unittest as ut
from io import StringIO
from model_bakery import baker
from pkg_resources import resource_filename
from django.test import TestCase
from django.core.management import call_command

from mkite.orm.structs.models import Crystal
from mkite.orm.jobs.models import Job, JobStatus, JobRecipe, Experiment


RULES_FILE = resource_filename("mkite.tests.files.workflow", "create_rules.yaml")


class TestCommand(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCommand, cls).setUpClass()

        for n in range(1, 6):
            baker.make(JobRecipe, name=f"test_recipe{n}")

        for n in range(1, 3):
            baker.make(Experiment, name=f"test_exp{n}")

    def call_command(self, *args, **kwargs):
        call_command(
            "create_from_file",
            *args,
            stdout=StringIO(),
            stderr=StringIO(),
            **kwargs,
        )

    def test_command(self):
        self.call_command("simple", RULES_FILE)

    def test_dry_run(self):
        self.call_command(
            "simple",
            RULES_FILE,
            "--dry_run",
        )



from model_bakery import baker
from django.test import TestCase

from mkite.orm.base.models import Formula, ChemNode, CalcNode
from mkite.orm.jobs.models import Job
from mkite.orm.jobs.serializers import JobSerializer
from mkite.orm.base.serializers import (
    FormulaSerializer,
    ChemNodeSerializer,
    CalcNodeSerializer,
)


class TestFormulaSerializer(TestCase):
    @property
    def dict(self):
        return {
            "name": "H20 C8 N1 +1",
            "charge": 1,
        }

    def test_serialize(self):
        f = baker.make(Formula, **self.dict)
        serial = FormulaSerializer(f)
        data = serial.data

        self.assertTrue("id" in data)
        expected = {
            "@module": "mkite.orm.base.models",
            "@class": "Formula",
            "name": "H20 C8 N1 +1",
            "charge": 1,
        }
        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        serial = FormulaSerializer(data=self.dict)

        self.assertTrue(serial.is_valid())

        new = serial.save()

        self.assertEqual(new.name, self.dict["name"])
        self.assertEqual(new.charge, self.dict["charge"])


class TestChemSerializer(TestCase):
    def test_deserialize(self):
        job = baker.make(Job)
        data = {
            "@module": "mkite.orm.base.models",
            "@class": "ChemNode",
            "parentjob": {"id": job.id},
        }
        serial = ChemNodeSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.parentjob, job)

    def test_serialize(self):
        node = baker.make(ChemNode)
        jobdata = JobSerializer(node.parentjob).data
        expected = {
            "@module": "mkite.orm.base.models",
            "@class": "ChemNode",
            "parentjob": jobdata,
            "id": node.id,
            "uuid": str(node.uuid),
        }
        serial = ChemNodeSerializer(node)

        self.assertEqual(serial.data, expected)


class TestCalcSerializer(TestCase):
    def test_deserialize(self):
        chem = baker.make(ChemNode)
        data = {
            "@module": "mkite.orm.base.models",
            "@class": "CalcNode",
            "parentjob": {"id": chem.parentjob.id},
            "chemnode": {"id": chem.id},
        }
        serial = CalcNodeSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.chemnode, chem)
        self.assertEqual(new.parentjob, chem.parentjob)

    def test_serialize(self):
        node = baker.make(CalcNode)
        jobdata = JobSerializer(node.parentjob).data
        chemdata = ChemNodeSerializer(node.chemnode).data
        expected = {
            "@module": "mkite.orm.base.models",
            "@class": "CalcNode",
            "parentjob": jobdata,
            "chemnode": chemdata,
            "id": node.id,
            "uuid": str(node.uuid),
        }
        serial = CalcNodeSerializer(node)

        self.assertEqual(serial.data, expected)



import os
from typing import Iterable, List
from mkite_core.external import load_config

from mkite_core.models import MoleculeInfo, ConformerInfo, NodeResults
from .base import DbImporter, DbImporterError


class MolFileImporter(DbImporter):
    RECIPE_DICT = {
        "name": "dbimport.MolFileImporter",
        "method": "EXT",
    }
    PACKAGE_DICT = {"name": "mkite.dbimport"}

    def query(self, **kwargs):
        """Opens the file provided and returns its contents.
        For now, does not perform any filtering."""

        if "filename" in kwargs:
            return self.read(kwargs["filename"])

        return [kwargs]

    def read(self, filename: os.PathLike):
        if filename.endswith(".json") or filename.endswith(".yaml"):
            return load_config(filename)

        if filename.endswith(".sdf"):
            raise NotImplementedError("Importing .sdf files is not yet supported")

    def convert(self, parsed: List[dict]) -> List[NodeResults]:
        """Converts the files into MoleculeInfo, which is then converted into a
        Node. The expected format of the dictionary of parsed information is
        the following:

        ```
            parsed[0] = {
                "smiles": "c1ccccc1",
                "attributes": {...},
            }
        ```
        """
        return [self.convert_item(item) for item in parsed]

    def convert_item(self, item: dict) -> NodeResults:
        info = MoleculeInfo.from_smiles(item["smiles"])
        if "attributes" in item:
            info.attributes = item["attributes"]

        return NodeResults(chemnode=info.as_dict())



import os
import shutil
import unittest as ut
from model_bakery import baker
from unittest.mock import patch
from django.test import TestCase

from mkite.orm.base.models import ChemNode
from mkite.orm.structs.models import Crystal
from mkite.orm.jobs.models import Job, JobStatus, JobRecipe, Experiment

from mkite.workflow.create import InputQuery
from mkite.workflow.create.simple import SimpleJobCreator


class TestJobCreator(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestJobCreator, cls).setUpClass()
        cls.inp_recipe = baker.make(JobRecipe)
        cls.out_recipe = baker.make(JobRecipe)
        cls.inp_experiment = baker.make(Experiment)
        cls.out_experiment = baker.make(Experiment)

        cls.chemnodes = baker.make(
            Crystal,
            _quantity=7,
            parentjob__recipe=cls.inp_recipe,
            parentjob__experiment=cls.inp_experiment,
            parentjob__status=JobStatus.DONE,
        )

        jobs = []
        for node in cls.chemnodes[:3]:
            job = baker.make(Job, recipe=cls.out_recipe, experiment=cls.out_experiment)
            job.inputs.add(node)
            jobs.append(job)

        cls.jobs = jobs

    @classmethod
    def tearDownClass(cls):
        pjobs = [node.parentjob for node in cls.chemnodes]

        for node in cls.chemnodes:
            node.delete()

        for j in cls.jobs + pjobs:
            j.delete()

        prj = cls.inp_experiment.project
        cls.inp_experiment.delete()
        prj.delete()

        prj = cls.out_experiment.project
        cls.out_experiment.delete()
        prj.delete()

        pkg = cls.out_recipe.package
        cls.out_recipe.delete()
        pkg.delete()

        pkg = cls.inp_recipe.package
        cls.inp_recipe.delete()
        pkg.delete()
        super(TestJobCreator, cls).tearDownClass()

    def setUp(self):
        inputs = [
            InputQuery(
                filter={
                    "parentjob__experiment__name": self.inp_experiment.name,
                    "parentjob__recipe__name": self.inp_recipe.name,
                },
            )
        ]

        self.creator = SimpleJobCreator(
            inputs,
            self.out_experiment.name,
            self.out_recipe.name,
            tags=["test_tag"],
        )

    def test_get_inputs(self):
        nodes = self.creator.get_inputs()
        self.assertEqual(nodes.count(), 4)
        self.assertFalse(self.chemnodes[-1] in nodes)

    def test_job_template(self):
        template = self.creator.job_template
        self.assertIsInstance(template, Job)

    def test_get_object(self):
        obj = self.creator.get_object(JobRecipe, id=self.inp_recipe.id)
        self.assertEqual(obj, self.inp_recipe)

    def test_create(self):
        nodes = self.creator.get_inputs()
        jobs, inputs = self.creator.create()

        self.assertEqual(len(jobs), 4)

        job = jobs[0]
        self.assertNotEqual(job.inputs.count(), 0)
        self.assertEqual(job.tags.count(), 1)
        tag = job.tags.first()
        self.assertEqual(str(tag), "test_tag")

        for j in jobs:
            j.delete()



import os
from django.db import models
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from mkite_engines import Status, EngineRoles, instantiate_from_path
from mkite.orm.jobs.models import Job, JobStatus, JobRecipe, Experiment, Project


class Command(BaseCommand):
    help = "Submits jobs using a given engine"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "engine_config",
            type=str,
            help="path to the configuration file for the engine that will be used to submit the jobs",
        )
        argparser.add_argument(
            "-r",
            "--recipe",
            type=str,
            default=None,
            help="Name of the recipe of the jobs to be submittted",
        )
        argparser.add_argument(
            "-p",
            "--project",
            type=str,
            default=None,
            help="Name of the project of the jobs to be submittted",
        )
        argparser.add_argument(
            "-e",
            "--experiment",
            type=str,
            default=None,
            help="Name of the experiment of the jobs to be submittted",
        )
        argparser.add_argument(
            "--dry_run",
            action="store_true",
            help="If set, does not submit anything",
        )
        return argparser

    def handle(self, engine_config, *args, dry_run=False, **kwargs):
        self.project = kwargs["project"]
        self.experiment = kwargs["experiment"]
        self.recipe = kwargs["recipe"]

        self.pub = instantiate_from_path(engine_config, role=EngineRoles.producer)

        jobs = self.get_jobs(**kwargs)

        num_jobs = jobs.count()
        if num_jobs == 0:
            self.log("error", "No jobs to submit. Exiting...")
            return

        if dry_run:
            self.log("success", f"(DRY_RUN) would have submitted {num_jobs} jobs.")
            return

        for job in jobs:
            self.submit_job(job)

        self.log("success", f"Submitted {num_jobs} jobs.")

    def get_jobs(self, **kwargs) -> models.QuerySet:
        self.log("notice", "Submitting jobs with the following constraints:")
        self.log("notice", f"Project: {self.project}")
        self.log("notice", f"Experiment: {self.experiment}")
        self.log("notice", f"Recipe: {self.recipe}")

        exp_args = self.get_experiment_args()
        rec_args = self.get_recipe_args()
        query = Job.objects.filter(
            status=JobStatus.READY,
            **exp_args,
            **rec_args,
        ).order_by("ctime")
        return query

    def get_experiment_args(self) -> dict:
        """Queries the experiments and returns the args of the query for
        the job. If the experiment is not specified, uses only the project
        name. If the project name is not specified, uses the experiment name.
        If neither are specified, simply submit all jobs."""
        try:
            if self.experiment is None and self.project is None:
                return {}

            elif self.experiment is None:
                prj = Project.objects.get(name=self.project)
                return {"experiment__project": prj.id}

            elif self.project is None:
                exp = Experiment.objects.get(name=self.experiment)
                return {"experiment": exp.id}

            else:
                exp = Experiment.objects.get(
                    name=self.experiment,
                    project__name=self.project,
                )
                return {"experiment": exp.id}

        except ObjectDoesNotExist:
            raise CommandError(
                f"Experiment {self.experiment} with project {self.project}\
                does not exist."
            )

    def get_recipe_args(self) -> dict:
        """Queries the recipes and returns the args of the query for
        the job. If the recipe is not specified, uses all recipes."""
        try:
            if self.recipe is None:
                return {}

            recipe = JobRecipe.objects.get(name=self.recipe)
            return {"recipe": recipe.id}

        except ObjectDoesNotExist:
            raise CommandError(f"Recipe {self.recipe} does not exist.")

    def submit_job(self, job: Job):
        info = job.as_info()
        recipe = job.recipe.name
        self.pub.push_info(recipe, info)
        job.status = JobStatus.RUNNING
        job.save()
        self.log("success", f"Submitted Job ID {job.id}.")



import json
from django.test import TestCase
from pkg_resources import resource_filename

from mkite.orm.deserializers import get_serializer
from mkite.orm.mols.models import Molecule
from mkite.orm.structs.models import Crystal


MOL_JSON = resource_filename("mkite_core.tests.files.models", "molecule.json")
CRYSTAL_JSON = resource_filename("mkite_core.tests.files.models", "crystal.json")


def load_json(filename):
    with open(filename, "r") as f:
        return json.load(f)


class TestDeserialization(TestCase):
    def setUp(self):
        self.mol_dict = load_json(MOL_JSON)
        self.crystal_dict = load_json(CRYSTAL_JSON)

    def test_mol(self):
        serial = get_serializer(self.mol_dict)
        serial.is_valid()
        self.assertTrue(serial.is_valid())

        obj = serial.save()

        self.assertIsInstance(obj, Molecule)

    def test_crystal(self):
        serial = get_serializer(self.crystal_dict)
        self.assertTrue(serial.is_valid())

        obj = serial.save()

        self.assertIsInstance(obj, Crystal)



import os
import itertools
from typing import List, Dict, Tuple

from django.db.models import QuerySet
from mkite.orm.base.models import ChemNode
from mkite.orm.jobs.models import Job

from .base import BaseJobCreator, JobCreationError


class TupleJobCreator(BaseJobCreator):
    def get_inputs(self) -> "List[QuerySet[ChemNode]]":
        inp_qs = self.get_input_queries()
        inp_to_exclude = self.get_inputs_with_jobs()
        inp_tuples = self.make_combinations(inp_qs, exclude=inp_to_exclude)

        return inp_tuples

    def get_existing_jobs(self) -> "QuerySet[Job]":
        return Job.objects.filter(
            experiment=self.out_experiment,
            recipe=self.out_recipe,
        )

    def get_inputs_with_jobs(self) -> List[Tuple[int]]:
        jobs = self.get_existing_jobs()
        job_inputs = jobs.values_list("id", "inputs")

        tuples = []
        for key, grp in itertools.groupby(job_inputs, key=lambda x: x[0]):
            inp_ids = tuple(sorted([x[1] for x in grp]))
            tuples.append(inp_ids)

        return tuples

    def get_input_queries(self) -> List[List[int]]:
        all_nodes = []
        for inpq in self.inputs:
            nodes = ChemNode.objects.filter(**inpq.filter)
            nodes = nodes.exclude(**inpq.exclude)
            nodes = list(nodes.values_list("id", flat=True))
            all_nodes.append(nodes)

        return all_nodes

    def make_combinations(
        self, query_sets: List[List[int]], exclude: List[Tuple[int]] = None
    ) -> List[QuerySet]:
        ninputs = len(query_sets)
        exclude = [] if exclude is None else exclude

        combinations = itertools.product(*query_sets)

        all_inputs = [
            ChemNode.objects.filter(id__in=tup)
            for tup in combinations
            if tuple(sorted(tup)) not in exclude
        ]

        return all_inputs

    def create(self, dry_run: bool = False) -> Tuple[List[Job], List[QuerySet]]:
        inputs = self.get_inputs()
        num_jobs = len(inputs)
        jobs = [self.job_template for n in range(num_jobs)]

        if dry_run:
            return jobs, inputs

        Job.objects.bulk_create(jobs, batch_size=self.batch_size)

        if self.tags:
            self.add_tags(jobs)

        JobParent = Job.inputs.through
        parents = [
            JobParent(chemnode_id=inp.id, job_id=job.id)
            for inp_set, job in zip(inputs, jobs)
            for inp in inp_set
        ]

        JobParent.objects.bulk_create(parents, batch_size=self.batch_size)

        return jobs, inputs



from model_bakery import baker
from django.test import TestCase

from mkite.orm.base.models import Formula
from mkite.orm.jobs.models import Job
from mkite.orm.mols.models import Molecule, Conformer


class MolCreator:
    def create_formula(self):
        formula, _ = Formula.objects.get_or_create(name="C9 H8 O4 +0", charge=0)
        return formula

    def create_molecule(self):
        ikey = "BSYNRYMUTXBXSQ-UHFFFAOYSA-N"
        mols = Molecule.objects.filter(inchikey=ikey)
        if mols.exists():
            return mols.first()

        mol = baker.make(
            Molecule,
            formula=self.create_formula(),
            inchikey=ikey,
            smiles="CC(=O)OC1=CC=CC=C1C(=O)O",
        )
        return mol

    def create_conformer(self):
        conf = baker.make(
            Conformer,
            formula=self.create_formula(),
            mol=self.create_molecule(),
            species=[
                "C",
                "C",
                "O",
                "O",
                "C",
                "C",
                "C",
                "C",
                "C",
                "C",
                "C",
                "O",
                "O",
                "H",
                "H",
                "H",
                "H",
                "H",
                "H",
                "H",
                "H",
            ],
        )
        return conf


class TestMols(TestCase):
    def setUp(self):
        self.creator = MolCreator()

    def test_molecule(self):
        mol = self.creator.create_molecule()

        self.assertTrue(hasattr(mol, "parentjob"))
        self.assertTrue(hasattr(mol, "chemnode_ptr"))
        self.assertEqual(mol.inchikey, "BSYNRYMUTXBXSQ-UHFFFAOYSA-N")
        self.assertEqual(mol.smiles, "CC(=O)OC1=CC=CC=C1C(=O)O")
        self.assertEqual(repr(mol), f"<Molecule: {mol.inchikey} ({mol.id})>")

    def test_formula(self):
        formula = self.creator.create_formula()
        self.assertEqual(formula.name, "C9 H8 O4 +0")

    def test_molecule_dict(self):
        mol = self.creator.create_molecule()
        data = mol.as_dict()

        expected = {
            "@module": "mkite.orm.mols.models",
            "@class": "Molecule",
            "id": mol.id,
            "uuid": str(mol.uuid),
            "ctime": mol.ctime,
            "mtime": mol.mtime,
            "parentjob": mol.parentjob.id,
            "chemnode_ptr": mol.chemnode_ptr.id,
            "attributes": mol.attributes,
            "siteprops": mol.siteprops,
            "formula": mol.formula.id,
            "inchikey": mol.inchikey,
            "smiles": mol.smiles,
            "tags": [],
        }

        self.assertEqual(data, expected)

        node = mol.chemnode_ptr
        self.assertEqual(node.get_data(), expected)

    def test_conformer(self):
        conf = self.creator.create_conformer()

        self.assertTrue(hasattr(conf, "parentjob"))
        self.assertTrue(hasattr(conf, "chemnode_ptr"))
        self.assertTrue(hasattr(conf, "mol"))
        self.assertEqual(len(conf.species), 21)
        self.assertEqual(
            repr(conf),
            f"<Conformer: {conf.formula.name}, Mol {conf.mol.inchikey} ({conf.id})>",
        )



from django.core.management.base import BaseCommand

from mkite.orm.jobs.models import Project


class Command(BaseCommand):
    help = "Adds a new project to the database"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "project",
            type=str,
            default=None,
            help="Name of the project that will be created",
        )
        argparser.add_argument(
            "--dry_run",
            action="store_true",
            help="If set, does not store objects into the database",
        )
        return argparser

    def handle(self, project, *args, dry_run=False, **kwargs):
        prj = Project.objects.filter(name=project)

        if prj.exists():
            self.log(
                "error",
                f"A Project named {project} already exists in the database.",
            )
            return

        if dry_run:
            self.log(
                "success",
                f"[DRY_RUN] Would have created Project {project}",
            )
        else:
            prj = Project.objects.create(name=project)
            self.log("success", f"Added project {project} (id {prj.id})")



import os
from typing import List
from mkite_core.external import load_config
from django.core.management.base import BaseCommand, CommandError

from mkite.workflow.create import InputQuery, JOB_CREATORS


class Command(BaseCommand):
    help = "Creates jobs according to the given input/output recipes"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "creator_name",
            type=str,
            choices=JOB_CREATORS.keys(),
            help="Type of JobCreator to use",
        )
        argparser.add_argument(
            "rules_file",
            type=str,
            help="File containing the rules on job creation",
        )
        argparser.add_argument(
            "-b",
            "--batch_size",
            type=int,
            default=None,
            help="Size of the batch when bulk creating new database objects",
        )
        argparser.add_argument(
            "--dry_run",
            action="store_true",
            help="If set, does not store anything into the database",
        )
        return argparser

    def handle(
        self, creator_name, rules_file, *args, batch_size=None, dry_run=False, **kwargs
    ):
        rules = self.get_rules(rules_file)
        creator_cls = JOB_CREATORS[creator_name]

        self.log("notice", f"File {rules_file}, creator {creator_name}")

        for i, r in enumerate(rules, 1):
            creator = creator_cls(
                inputs=r["inputs"],
                out_experiment=r["out_experiment"],
                out_recipe=r["out_recipe"],
                options=r.get("options", None),
                tags=r.get("tags", None),
                batch_size=batch_size,
            )

            self.log("notice", f"Rule {i}: ({r['out_experiment']}, {r['out_recipe']})")

            jobs, inputs = creator.create(dry_run=dry_run)

            msg = f"created {len(jobs)} new jobs."
            if dry_run:
                msg = "(DRY_RUN) would have " + msg

            self.log("success", msg)

    def get_rules(self, rules_file: os.PathLike) -> List[dict]:
        """Loads the rules from the file. Expected format (in Python dictionary):

        ```
        rules = [
            {
                "inp_experiment": "Experiment1",
                "inp_recipe": "Recipe1",
                "out_experiment": "Experiment2",
                "out_recipe": "Recipe2",
                "options": {...},
                "tags": ["tag3"],
                "filter_kwargs": {
                    "parentjob__tags__in": ["tag1", "tag2"],
                },
                "exclude_kwargs": {},
            },
            ...
        ]
        ```
        """

        rules = load_config(rules_file)
        if isinstance(rules, dict):
            rules = [rules]

        validated_rules = [r for r in rules if self.is_valid_rule(r)]

        return validated_rules

    def is_valid_rule(self, rule: dict) -> bool:
        if not isinstance(rule, dict):
            return False

        required_keys = ["inputs", "out_experiment", "out_recipe"]
        return all([k in rule for k in required_keys])



from model_bakery import baker
from django.test import TestCase

from mkite.orm.base.models import Formula
from mkite.orm.jobs.models import Job
from mkite.orm.structs.models import Crystal, SpaceGroups
from mkite_core.models import CrystalInfo


class CrystalCreator:
    def create_formula(self):
        formula, _ = Formula.objects.get_or_create(
            name="Si2 +0",
            charge=0,
        )
        return formula

    def create_crystal(self):
        crystal = baker.make(
            Crystal,
            formula=self.create_formula(),
            spacegroup=SpaceGroups(227),
            species=["Si", "Si"],
            coords=[[0.0, 0.0, 0.0], [1.365, 1.365, 1.365]],
            lattice=[[0.0, 2.73, 2.73], [2.73, 0.0, 2.73], [2.73, 2.73, 0.0]],
        )
        return crystal


class TestStructs(TestCase):
    def setUp(self):
        self.creator = CrystalCreator()

    def test_crystal(self):
        crystal = self.creator.create_crystal()

        self.assertTrue(hasattr(crystal, "parentjob"))
        self.assertTrue(hasattr(crystal, "chemnode_ptr"))
        self.assertTrue(hasattr(crystal, "tags"))

        self.assertEqual(crystal.spacegroup, 227)
        self.assertEqual(crystal.species, ["Si", "Si"])

        info = crystal.as_info()
        self.assertIsInstance(info, CrystalInfo)

    def test_formula(self):
        formula = self.creator.create_formula()
        self.assertEqual(formula.name, "Si2 +0")
        self.assertEqual(formula.charge, 0)



import os
import environ
from pathlib import Path

MKITE_ENV_KEY = "MKITE_ENV"
if not MKITE_ENV_KEY in os.environ:
    raise ValueError(
        f"{MKITE_ENV_KEY} is not defined in the environment. \
        Without this variable, it is unclear what database \
        or settings to use. Please create a .env file and \
        point the {MKITE_ENV_KEY} file to it."
    )
MKITE_ENV = os.environ[MKITE_ENV_KEY]
if not os.path.exists(MKITE_ENV):
    raise FileNotFoundError(
        f"Could not find environment file {MKITE_ENV}."
    )

# Reads the access to the database via environmental variables
env = environ.Env(DEBUG=(bool, False))
env.read_env(MKITE_ENV)

BASE_DIR = Path(__file__).resolve().parent.parent
DEBUG = env("DEBUG")
SECRET_KEY = env("SECRET_KEY")
ALLOWED_HOSTS = env.list("DJANGO_ALLOWED_HOSTS", default=["*"])

# Sets up the database connection
DATABASES = {'default': env.db()}

# Application definition
DJANGO_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

THIRD_PARTY_APPS = ["django_extensions", "rest_framework", "taggit"]

LOCAL_APPS = [
    "mkite.orm.base",
    "mkite.orm.jobs",
    "mkite.orm.calcs",
    "mkite.orm.mols",
    "mkite.orm.structs",
    "mkite.workflow",
]

INSTALLED_APPS = DJANGO_APPS + THIRD_PARTY_APPS + LOCAL_APPS

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "cfg.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

WSGI_APPLICATION = "cfg.wsgi.application"


# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_TZ = True


# Static files (CSS, JavaScript, Images)
STATIC_URL = "static/"

# Default primary key field type
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



from importlib import metadata
from django.core.management.base import BaseCommand, CommandError

from mkite_core.plugins import get_recipes
from mkite.orm.jobs.models import JobRecipe
from mkite.orm.jobs.serializers import JobRecipeSerializer


class Command(BaseCommand):
    help = "Creates jobs according to the given input/output recipes"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "--dry_run",
            action="store_true",
            help="If set, does not store objects into the database",
        )
        return argparser

    def handle(self, *args, dry_run=False, **kwargs):
        self.log("notice", f"Scanning entry point mkite.recipes...")
        recipes = get_recipes()

        new_recipes = []
        for name, entry in recipes.items():
            new = self.add_entry_point(name, entry, dry_run=dry_run)
            if new is not None:
                new_recipes.append(new)

        if dry_run:
            self.log("success", f"Found {len(new_recipes)} recipes in entry points")
        else:
            self.log("success", f"Added {len(new_recipes)} recipes to the database")

    def add_entry_point(
        self, name: str, entry: metadata.EntryPoint, dry_run: bool = False
    ):
        cls = entry.load()
        data = {
            "name": name,
            "method": cls._METHOD,
            "package": {"name": cls._PACKAGE_NAME},
        }

        if dry_run:
            self.log("success", f"Found recipe {name} in entry points")
            return data

        try:
            new = self.create_recipe(data)
            self.log("success", f"Added recipe {name} to the database")
            return new

        except Exception as e:
            self.log("error", f"Error adding recipe {name} to the database: {str(e)}")
            return None

    def create_recipe(self, data: dict) -> JobRecipe:
        serial = JobRecipeSerializer(data=data)
        if not serial.is_valid():
            raise CommandError(f"Error deserializing data: {serial.errors}")

        return serial.save()



import json
import unittest as ut
from io import StringIO
from model_bakery import baker
from django.test import TestCase
from django.core.management import call_command

from mkite.orm.structs.models import Crystal
from mkite.orm.jobs.models import Job, JobStatus, JobRecipe, Experiment


class TestCommand(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCommand, cls).setUpClass()
        cls.inp_recipe = baker.make(JobRecipe)
        cls.out_recipe = baker.make(JobRecipe)
        cls.inp_experiment = baker.make(Experiment)
        cls.out_experiment = baker.make(Experiment)

        cls.chemnodes = baker.make(
            Crystal,
            _quantity=7,
            parentjob__recipe=cls.inp_recipe,
            parentjob__experiment=cls.inp_experiment,
            parentjob__status=JobStatus.DONE,
        )

        jobs = []
        for node in cls.chemnodes[:3]:
            job = baker.make(Job, recipe=cls.out_recipe, experiment=cls.out_experiment)
            job.inputs.add(node)
            jobs.append(job)

        cls.jobs = jobs

    def call_command(self, *args, **kwargs):
        call_command(
            "create_simple",
            *args,
            stdout=StringIO(),
            stderr=StringIO(),
            **kwargs,
        )

    def test_command(self):
        opts = json.dumps({"test": 1})

        query = Job.objects.filter(
            experiment=self.out_experiment,
            recipe=self.out_recipe,
            status=JobStatus.READY,
        )
        self.assertEqual(query.count(), 3)

        tag = "test_tag1"
        self.call_command(
            self.inp_experiment.name,
            self.inp_recipe.name,
            self.out_experiment.name,
            self.out_recipe.name,
            "--options",
            opts,
            "--tags",
            tag,
        )

        query = Job.objects.filter(
            recipe=self.out_recipe,
            status=JobStatus.READY,
            tags__name=tag,
        )
        self.assertEqual(query.count(), 4)

        query = Job.objects.filter(
            recipe=self.out_recipe,
            status=JobStatus.READY,
        )
        self.assertEqual(query.count(), 7)



from rest_framework import serializers

from mkite.orm.serializers import BaseSerializer
from mkite.orm.jobs.serializers import JobSerializer
from .models import Formula, ChemNode, CalcNode


class FormulaSerializer(BaseSerializer):
    class Meta:
        model = Formula
        fields = "__all__"


class ChemNodeSerializer(BaseSerializer):
    parentjob = JobSerializer(nested_field=True)

    class Meta:
        model = ChemNode
        fields = (
            "id",
            "uuid",
            "parentjob",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )


class CalcNodeSerializer(BaseSerializer):
    parentjob = JobSerializer(nested_field=True)
    chemnode = ChemNodeSerializer(nested_field=True)

    class Meta:
        model = CalcNode
        fields = (
            "id",
            "uuid",
            "parentjob",
            "chemnode",
        )
        read_only_fields = ("ctime", "mtime")



import os
from typing import Iterable
from django.core.management.base import BaseCommand, CommandError

from mkite_core.models import JobResults
from mkite.workflow.parse import JobParser
from mkite_engines import Status, EngineRoles, instantiate_from_path

from mkite.orm.jobs.models import Job


class Command(BaseCommand):
    help = "Parses the job results from a folder into the database"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "filename",
            type=str,
            help="path to a file to be parsed. If given, ignores the choice of engine and parses only that file",
        )
        return argparser

    def handle(self, filename, **kwargs):
        self.log("notice", f"Parsing file: {filename}")
        results = JobResults.from_json(filename)
        self.parse_result(results)

    def parse_result(self, info: JobResults) -> bool:
        jobstr = self.get_info_string(info)
        try:
            if not self.is_valid_parse(info):
                raise CommandError(
                    f"Invalid parsing of {jobstr}. Job is likely done already."
                )

            parser = JobParser(info)
            out = parser.parse()
            self.log("success", f"Parsed {jobstr}")

            return out

        except Exception as e:
            self.log("error", f"Error processing {jobstr}: {str(e)}")

            return None

    def is_valid_parse(self, info: JobResults):
        """Verifies if it is valid to parse the JobResults given by info. This prevents
        duplicate jobs from being parsed into the system.
        """
        if "id" not in info.job and "uuid" not in info.job:
            return True

        if "id" in info.job:
            job = Job.objects.get(id=info.job["id"])

        elif "uuid" in info.job:
            job = Job.objects.get(uuid=info.job["uuid"])

        return job.status != "D"

    def get_dict_string(self, data: dict, prefix="job"):
        if not prefix.endswith(" "):
            prefix += " "

        if "uuid" in data:
            return prefix + data["uuid"]

        elif "id" in data:
            return prefix + str(data["id"])

        elif "name" in data:
            return prefix + data["name"]

        return ""

    def get_info_string(self, info: JobResults):
        string = self.get_dict_string(info.job, prefix="job")

        if string:
            return string

        elif "experiment" in info.job:
            string = self.get_dict_string(info.job["experiment"], prefix="experiment")
            if string:
                return string

        elif "recipe" in info.job:
            string = self.get_dict_string(info.job["recipe"], prefix="recipe")
            if string:
                return string

        return ""



import unittest as ut
from datetime import timedelta
from model_bakery import baker
from django.test import TestCase
from rest_framework import serializers
from collections.abc import Mapping

from mkite.orm.jobs.models import (
    Project,
    Experiment,
    JobPackage,
    JobRecipe,
    RunStats,
    Job,
    JobStatus,
)
from mkite.orm.jobs.serializers import (
    ProjectSerializer,
    ExperimentSerializer,
    JobPackageSerializer,
    JobRecipeSerializer,
    RunStatsSerializer,
    JobSerializer,
)

from .test_models import JobTestCreator


class SerializerTester(TestCase):
    def setUp(self):
        self.creator = JobTestCreator()


class TestProjectSerializer(SerializerTester):
    def test_serialize(self):
        obj = self.creator.create_project()
        serial = ProjectSerializer(obj)
        data = serial.data

        expected = {
            "id": obj.id,
            "uuid": str(obj.uuid),
            "name": "test_prj",
            "description": "project for testing purposes",
        }

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        data = {
            "name": "test_prj_2",
            "description": "testing",
        }
        serial = ProjectSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.name, "test_prj_2")
        self.assertEqual(new.description, "testing")


class TestExperimentSerializer(SerializerTester):
    def test_serialize(self):
        obj = self.creator.create_experiment()
        serial = ExperimentSerializer(obj)
        data = serial.data

        expected = {
            "id": obj.id,
            "uuid": str(obj.uuid),
            "name": "test_exp",
            "description": "first experiment",
        }

        self.assertTrue("project" in data)
        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        prj = baker.make(Project)
        data = {
            "project": {"id": prj.id},
            "name": "test_exp_2",
            "description": "testing",
        }
        serial = ExperimentSerializer(data=data)
        serial.is_valid()
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.project.name, prj.name)
        self.assertEqual(new.name, "test_exp_2")
        self.assertEqual(new.description, "testing")


class TestJobPackageSerializer(SerializerTester):
    def test_serialize(self):
        obj = self.creator.create_package()
        serial = JobPackageSerializer(obj)
        data = serial.data

        expected = {
            "id": obj.id,
            "uuid": str(obj.uuid),
            "name": obj.name,
        }

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        data = {
            "name": "test_pkg_2",
        }
        serial = JobPackageSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in data.items():
            self.assertEqual(getattr(new, k), v)


class TestJobRecipeSerializer(SerializerTester):
    def test_serialize(self):
        obj = self.creator.create_recipe()
        serial = JobRecipeSerializer(obj)
        data = serial.data

        expected = {
            "id": obj.id,
            "uuid": str(obj.uuid),
            "name": obj.name,
            "method": obj.method,
        }

        self.assertTrue("package" in data)
        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        pkg = baker.make(JobPackage)
        data = {
            "package": {"uuid": pkg.uuid},
            "name": "test_recipe_2",
            "method": "FF",
        }
        serial = JobRecipeSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in data.items():
            if not isinstance(v, Mapping):
                self.assertEqual(getattr(new, k), v)


class TestRunStatsSerializer(SerializerTester):
    def test_serialize(self):
        obj = self.creator.create_runstats()
        serial = RunStatsSerializer(obj)
        data = serial.data

        expected = {
            "id": obj.id,
            "uuid": str(obj.uuid),
            "host": obj.host,
            "cluster": obj.cluster,
            "duration": obj.duration.total_seconds(),
            "ncores": obj.ncores,
            "ngpus": obj.ngpus,
            "pkgversion": obj.pkgversion,
        }

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        data = {
            "host": "test_host_2",
            "cluster": "test_cluster_2",
            "duration": timedelta(seconds=1000),
            "ncores": 6,
            "ngpus": 1,
            "pkgversion": "1.0",
        }
        serial = RunStatsSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in data.items():
            self.assertEqual(getattr(new, k), v)

    def test_deserialize_duration_float(self):
        data = {
            "host": "test_host_2",
            "cluster": "test_cluster_2",
            "duration": 0.03,
            "ncores": 6,
            "ngpus": 1,
            "pkgversion": "1.0",
        }
        serial = RunStatsSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in data.items():
            if k == "duration":
                continue

            self.assertEqual(getattr(new, k), v)

    def test_deserialize_duration_decimal(self):
        data = {
            "host": "test_host_2",
            "cluster": "test_cluster_2",
            "duration": 0.5329999990000000000003,
            "ncores": 6,
            "ngpus": 1,
            "pkgversion": "1.0",
        }
        serial = RunStatsSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in data.items():
            if k == "duration":
                continue

            self.assertEqual(getattr(new, k), v)


class TestJobSerializer(SerializerTester):
    def test_serialize(self):
        obj = self.creator.create_job()
        serial = JobSerializer(obj)
        data = serial.data

        expected = {
            "id": obj.id,
            "uuid": str(obj.uuid),
            "status": obj.status,
            "isroot": obj.isroot,
        }

        for k in ["experiment", "recipe", "runstats", "tags"]:
            self.assertTrue(k in data)

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        recipe = baker.make(JobRecipe)
        experiment = baker.make(Experiment)
        runstats_data = {
            "host": "test_host_2",
            "cluster": "test_cluster_2",
            "duration": timedelta(seconds=1000),
            "ncores": 6,
            "ngpus": 1,
        }
        job_data = {
            "experiment": {"id": experiment.id},
            "recipe": {"name": recipe.name},
            "runstats": runstats_data,
            "status": JobStatus.DONE,
            "tags": ["tag_1", "tag_2"],
        }

        serial = JobSerializer(data=job_data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in job_data.items():
            if not isinstance(v, Mapping) and k != "tags":
                self.assertEqual(getattr(new, k), v)

    def test_invalid_runstats(self):
        recipe = baker.make(JobRecipe)
        experiment = baker.make(Experiment)
        invalid_runstats_data = {
            "ncores": 6,
            "ngpus": 1,
        }
        job_data = {
            "experiment": {"id": experiment.id},
            "recipe": {"id": recipe.id},
            "runstats": invalid_runstats_data,
            "status": JobStatus.DONE,
        }
        serial = JobSerializer(data=job_data)
        self.assertTrue(serial.is_valid())
        with self.assertRaises(serializers.ValidationError):
            serial.save()

    def test_deserialize_new_job(self):
        recipe = baker.make(JobRecipe)
        experiment = baker.make(Experiment)
        job_data = {
            "experiment": {"id": experiment.id},
            "recipe": {"name": recipe.name},
            "status": JobStatus.READY,
        }

        serial = JobSerializer(data=job_data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        for k, v in job_data.items():
            if not isinstance(v, Mapping):
                self.assertEqual(getattr(new, k), v)



import os
import json
import shutil
import unittest as ut
from unittest.mock import patch
from io import StringIO
from model_bakery import baker
from django.test import TestCase, SimpleTestCase
from django.core.management import call_command

from pkg_resources import resource_filename

from mkite.orm.jobs.models import Job, Experiment, Project
from mkite_core.models import JobResults
from mkite_core.tests.tempdirs import run_in_tempdir
from mkite.workflow.management.commands.dbimport import Command, DB_IMPORTERS

from mkite.dbimport.tests.test_mp import MockMPImporter


JOB_RESULTS_FILE = resource_filename("mkite.tests.files.workflow", "jobresults.json")
MP_QUERY_FILE = resource_filename("mkite.tests.files.dbimport", "mp_query.json")

MOCK_DB_IMPORTERS = {
    "MockMPImporter": MockMPImporter,
}


class QueryMixin:
    def get_query(self):
        with open(MP_QUERY_FILE, "r") as f:
            data = json.load(f)

        return data

    def get_query_string(self):
        return json.dumps(self.get_query()[0])


class CommandSimpleTests(SimpleTestCase, QueryMixin):
    def setUp(self):
        self.cmd = Command()

    def get_kwargs(self):
        return {
            "project": "test_dbimport",
            "experiment": "test",
        }

    def test_db_importers(self):
        self.assertTrue("MPImporter" in DB_IMPORTERS)

    def test_query_args(self):
        kwargs = {
            **self.get_kwargs(),
            "query": self.get_query_string(),
        }
        query = self.cmd.get_query_args(**kwargs)
        self.assertEqual(query, self.get_query()[0])

        kwargs = {
            **self.get_kwargs(),
            "file": MP_QUERY_FILE,
        }
        query = self.cmd.get_query_args(**kwargs)
        self.assertEqual(query, self.get_query())


class TestCommand(TestCase, QueryMixin):
    def call_command(self, *args, **kwargs):
        call_command(
            "dbimport",
            *args,
            stdout=StringIO(),
            stderr=StringIO(),
            **kwargs,
        )

    def get_command(self):
        return Command(stdout=StringIO(), stderr=StringIO())

    def get_info(self):
        return JobResults.from_json(JOB_RESULTS_FILE)

    def test_is_valid_parse(self):
        cmd = self.get_command()
        info = self.get_info()
        info.job["options"]["test"] = "test_options"

        cmd.project = info.job["experiment"]["project"]["name"]
        cmd.experiment = info.job["experiment"]["name"]

        self.assertTrue(cmd.is_valid_parse(info))

        job = baker.make(
            Job,
            project__name=cmd.project,
            experiment__name=cmd.experiment,
            options=info.job["options"],
        )

        self.assertFalse(cmd.is_valid_parse(info))
        job.delete()

    def test_save_jobresults(self):
        cmd = self.get_command()
        info = self.get_info()
        cmd.project = info.job["experiment"]["project"]["name"]
        cmd.experiment = info.job["experiment"]["name"]

        out = cmd.save_jobresults(info)

        self.assertTrue(out is not None)
        self.assertIsInstance(out.job, Job)
        self.assertTrue(hasattr(out.job, "id"))

    @ut.skipIf("MP_API_KEY" not in os.environ, "MP_API_KEY is not in environment")
    @patch.dict(
        "mkite.workflow.management.commands.dbimport.DB_IMPORTERS", MOCK_DB_IMPORTERS
    )
    def test_call(self):
        self.call_command(
            "MockMPImporter",
            "--project",
            "test_dbimport",
            "--experiment",
            "test_exp",
            "--query",
            self.get_query_string(),
        )



import os
import sys
from pathlib import Path
from subprocess import call


def main():
    manage_py_path = Path(__file__).resolve().parent.parent / "manage.py"

    if not manage_py_path.is_file():
        print("Error: manage.py not found.")
        sys.exit(1)

    sys.argv.pop(0)  # Remove the script name from the arguments
    call([sys.executable, str(manage_py_path)] + sys.argv)


if __name__ == "__main__":
    main()



from typing import List, Dict, Union, Iterable, Tuple
from abc import ABC, abstractmethod
from pydantic import BaseModel, Field

from mkite.orm.base.models import ChemNode
from mkite.orm.jobs.models import Job, JobRecipe, Experiment


class JobCreationError(Exception):
    pass


class InputQuery(BaseModel):
    """Dictionary containing all the information to filter/exclude
    when querying what are going to be the inputs of the job.
    The input dictionary should be of the following format:

    ```
    "filter": {
        "parentjob__experiment__name": "test_exp",
        "parentjob__recipe__name": "test_recipe",
        "parentjob__tags__name": "tag",
        "crystal__isnull": True,
    }
    "exclude": {
        "parentjob__tags__name": "tag2",
    }
    ```
    """

    filter: dict = Field(
        default_factory=dict,
        description="kwargs used to filter nodes. Anything accepted by django",
    )
    exclude: dict = Field(
        default_factory=dict,
        description="kwargs used to exclude nodes. Anything accepted by django",
    )


class BaseJobCreator(ABC):
    def __init__(
        self,
        inputs: List[InputQuery],
        out_experiment: str,
        out_recipe: str,
        options: dict = None,
        tags: List[str] = None,
        batch_size: int = None,
    ):
        """Initializes the JobCreator with the information required to
        create all jobs.

        Args:
            inputs: dictionary containing all the information to filter/exclude
                when querying what are going to be the inputs of the job. As
                the job may take more than one input, it requires a list of
                dictionaries. Used according to the format of `InputQuery`.
            out_experiment: name of the experiment of the job to be created. Should
                already exist in the database.
            out_recipe: name of the recipe of the job to be created. Should already
                exist in the database.
            options: dicionary containing the special options to override the defaults
                in the job.
            tags: list of tags to add to the new job
            batch_size: size of the batch when bulk creating new jobs.
        """

        self.inputs = self.format_inputs(inputs)
        self.out_experiment = self.get_object(model=Experiment, name=out_experiment)
        self.out_recipe = self.get_object(model=JobRecipe, name=out_recipe)

        self.options = options if options is not None else {}
        self.tags = self.format_tags(tags)
        self.batch_size = batch_size

    def format_inputs(self, inputs: List[Union[dict, InputQuery]]) -> List[InputQuery]:
        formatted = []
        for inp in inputs:
            if isinstance(inp, InputQuery):
                formatted.append(inp)

            elif isinstance(inp, dict):
                iq = InputQuery(**inp)
                formatted.append(iq)

            else:
                raise JobCreationError(f"Invalid input type {type(inp)}")

        return formatted

    def get_object(self, model, **kwargs):
        return model.objects.get(**kwargs)

    @abstractmethod
    def get_inputs(self) -> Iterable[ChemNode]:
        """Gets the inputs of the jobs"""

    @property
    def job_template(self) -> Job:
        return Job(
            experiment=self.out_experiment,
            recipe=self.out_recipe,
            options=self.options,
            isroot=False,
        )

    def format_tags(self, tags) -> List[str]:
        if tags is None:
            return []

        if type(tags) == str:
            return [tags]

        return tags

    def add_tags(self, jobs: List[Job]):
        for j in jobs:
            j.tags.add(*self.tags)

    @abstractmethod
    def create(self, dry_run: bool = False) -> Tuple[List[Job], Iterable]:
        """Creates the jobs based on the inputs"""



from django.db import transaction
from rest_framework import serializers

from taggit.serializers import TagListSerializerField, TaggitSerializer
from mkite.orm.serializers import BaseSerializer
from mkite.orm.base.models import Formula
from mkite.orm.base.serializers import FormulaSerializer, ChemNodeSerializer

from .models import Molecule, Conformer


class MoleculeSerializer(TaggitSerializer, ChemNodeSerializer):
    formula = FormulaSerializer(nested_field=True, required=False)
    tags = TagListSerializerField(required=False)

    class Meta:
        model = Molecule
        fields = "__all__"
        read_only_fields = ("inchikey",)

    @transaction.atomic
    def create(self, validated_data):
        from mkite_core.external.rdkit import RdkitInterface

        iface = RdkitInterface.from_smiles(validated_data["smiles"])

        validated_data.update(
            {
                "inchikey": iface.inchikey,
                "smiles": iface.smiles,
                "formula": {"name": iface.formula, "charge": iface.charge},
            }
        )

        return super().create(validated_data)


class ConformerSerializer(ChemNodeSerializer):
    formula = FormulaSerializer(nested_field=True, required=False)
    mol = MoleculeSerializer(nested_field=True, required=False)

    class Meta:
        model = Conformer
        fields = "__all__"



import unittest as ut
from pkg_resources import resource_filename

from ase import Atoms
from ase.io import read
from mkite_core.models import CrystalInfo, ConformerInfo, NodeResults
from mkite.dbimport.ase import AseFileImporter


CIF_FILE = resource_filename("mkite.tests.files.dbimport", "silicon.cif")
XYZ_FILE = resource_filename("mkite.tests.files.dbimport", "caffeine.xyz")


class TestAseFileImporter(ut.TestCase):
    def setUp(self):
        self.dbimp = AseFileImporter(
            project="test_prj",
            experiment="test_exp",
        )

    def get_crystal(self):
        return read(CIF_FILE)

    def get_conformer(self):
        return read(XYZ_FILE)

    def test_read_cif(self):
        data = self.dbimp.read(CIF_FILE)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 1)
        self.assertIsInstance(data[0], Atoms)

    def test_read_xyz(self):
        data = self.dbimp.read(XYZ_FILE)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 1)
        self.assertIsInstance(data[0], Atoms)

    def test_query(self):
        data = self.dbimp.query(CIF_FILE)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 1)
        self.assertIsInstance(data[0], Atoms)

    def test_convert_crystal(self):
        item = self.get_crystal()

        results = self.dbimp.convert_item(item)

        self.assertIsInstance(results, NodeResults)
        self.assertIsInstance(results.chemnode, dict)

        info = CrystalInfo.from_dict(results.chemnode)

    def test_convert_conformer(self):
        item = self.get_conformer()

        results = self.dbimp.convert_item(item)

        self.assertIsInstance(results, NodeResults)
        self.assertIsInstance(results.chemnode, dict)

        info = ConformerInfo.from_dict(results.chemnode)

    def test_convert(self):
        data = [
            self.get_crystal(),
            self.get_conformer(),
        ]
        results = self.dbimp.convert(data)

        self.assertIsInstance(results, list)
        self.assertIsInstance(results[0], NodeResults)
        self.assertEqual(len(results), 2)



import os
import shutil
from io import StringIO
from django.test import TestCase
from django.core.management import call_command
from pkg_resources import resource_filename

from mkite.orm.jobs.models import Job
from mkite_core.models import JobResults
from mkite_core.external import load_config
from mkite_core.tests.tempdirs import run_in_tempdir
from mkite_engines import Status
from mkite.workflow.management.commands.parse import Command


ENGINE = resource_filename("mkite.tests.files", "engine.yaml")
ENGINE_CFG = load_config(ENGINE)
JOB_RESULTS_FILE = resource_filename("mkite.tests.files.workflow", "jobresults.json")


def _prepare_folder():
    path = os.path.join(ENGINE_CFG["root_path"], Status.PARSING.value)
    os.mkdir(path)
    shutil.copy(JOB_RESULTS_FILE, path)
    return path


class TestParserCommand(TestCase):
    def call_command(self, *args, **kwargs):
        call_command(
            "parse_file",
            *args,
            stdout=StringIO(),
            stderr=StringIO(),
            **kwargs,
        )

    def get_command(self):
        return Command(stdout=StringIO(), stderr=StringIO())

    def get_info(self):
        return JobResults.from_json(JOB_RESULTS_FILE)

    def test_is_valid_parse(self):
        cmd = self.get_command()
        info = self.get_info()

        self.assertTrue(cmd.is_valid_parse(info))

    def test_parse_result(self):
        cmd = self.get_command()
        info = self.get_info()
        out = cmd.parse_result(info)

        self.assertTrue(out is not None)
        self.assertIsInstance(out.job, Job)
        self.assertTrue(hasattr(out.job, "id"))

    @run_in_tempdir
    def test_call(self):
        _prepare_folder()
        self.call_command(JOB_RESULTS_FILE)

        info = self.get_info()
        query = Job.objects.filter(
            experiment__name=info.job["experiment"]["name"],
            recipe__name=info.job["recipe"]["name"],
            runstats__cluster=info.runstats.cluster,
        )
        self.assertTrue(query.exists())

    @run_in_tempdir
    def test_call_filename(self):
        self.call_command(JOB_RESULTS_FILE)

        info = self.get_info()
        query = Job.objects.filter(
            experiment__name=info.job["experiment"]["name"],
            recipe__name=info.job["recipe"]["name"],
            runstats__cluster=info.runstats.cluster,
        )
        self.assertTrue(query.exists())



import unittest as ut
from django.test import TestCase
from model_bakery import baker

from mkite.orm.models import Crystal, Job, RunStats
from mkite.orm.jobs.deleter import delete_tree


class TestDeleter(TestCase):
    def setUp(self):
        """Creates a tree of jobs"""
        self.node1 = baker.make(Crystal)
        self.node2 = baker.make(Crystal)
        self.node2.parentjob.inputs.add(self.node1)
        self.node3 = baker.make(Crystal, parentjob=self.node2.parentjob)

        self.node4 = baker.make(Crystal)
        self.node4.parentjob.inputs.add(self.node2)

        self.node5 = baker.make(Crystal)
        self.node5.parentjob.inputs.add(self.node3)

        self.leaf_job = baker.make(Job)
        self.leaf_stats = baker.make(RunStats)
        self.leaf_job.inputs.add(self.node5)
        self.leaf_job.runstats = self.leaf_stats
        self.leaf_job.save()

    def test_delete_leaf(self):
        delete_tree(self.leaf_job)
        self.assertIsNone(self.leaf_job.pk)
        self.assertIsNone(self.leaf_stats.pk)
        self.assertFalse(self.node5.childjobs.exists())

    def test_delete_node5(self):
        delete_tree(self.node5.parentjob)
        node5q = Crystal.objects.filter(id=self.node5.id)
        self.assertFalse(node5q.exists())
        self.assertFalse(self.node3.childjobs.exists())

    def test_delete_node2(self):
        delete_tree(self.node2.parentjob)
        node2q = Crystal.objects.filter(id=self.node2.id)
        self.assertFalse(node2q.exists())

    def test_delete_node1(self):
        delete_tree(self.node1.parentjob)
        qs = Crystal.objects.all()
        self.assertEqual(qs.count(), 0)
        self.assertEqual(Job.objects.count(), 0)



import unittest as ut
from model_bakery import baker
from unittest.mock import patch
from django.test import TestCase

from mkite.orm.base.models import ChemNode
from mkite.orm.structs.models import Crystal
from mkite.orm.jobs.models import Job, JobStatus, JobRecipe, Experiment

from mkite.workflow.create import InputQuery
from mkite.workflow.create.tuple import TupleJobCreator


class TestJobCreator(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestJobCreator, cls).setUpClass()
        cls.recipe_1 = baker.make(JobRecipe)
        cls.recipe_2 = baker.make(JobRecipe)
        cls.out_recipe = baker.make(JobRecipe)
        cls.inp_experiment = baker.make(Experiment)
        cls.out_experiment = baker.make(Experiment)

        cls.crystals = baker.make(
            Crystal,
            _quantity=7,
            parentjob__recipe=cls.recipe_1,
            parentjob__experiment=cls.inp_experiment,
            parentjob__status=JobStatus.DONE,
        )

        cls.conformers = baker.make(
            Crystal,
            _quantity=5,
            parentjob__recipe=cls.recipe_2,
            parentjob__experiment=cls.inp_experiment,
            parentjob__status=JobStatus.DONE,
        )

        NJOBS = 3
        jobs = []
        for crystal, conf in zip(cls.crystals[:NJOBS], cls.conformers[:NJOBS]):
            job = baker.make(Job, recipe=cls.out_recipe, experiment=cls.out_experiment)
            job.inputs.add(crystal)
            job.inputs.add(conf)
            jobs.append(job)

        cls.jobs = jobs

    @classmethod
    def tearDownClass(cls):
        pjobs = [node.parentjob for node in cls.crystals]
        pjobs += [node.parentjob for node in cls.conformers]

        for j in cls.jobs + pjobs:
            j.delete()

        for exp in [cls.inp_experiment, cls.out_experiment]:
            prj = exp.project
            exp.delete()
            prj.delete()

        for recipe in [cls.recipe_1, cls.recipe_2, cls.out_recipe]:
            pkg = recipe.package
            recipe.delete()
            pkg.delete()

        super(TestJobCreator, cls).tearDownClass()

    def setUp(self):
        inputs = [
            InputQuery(
                filter={
                    "parentjob__experiment__name": self.inp_experiment.name,
                    "parentjob__recipe__name": self.recipe_1.name,
                },
            ),
            InputQuery(
                filter={
                    "parentjob__experiment__name": self.inp_experiment.name,
                    "parentjob__recipe__name": self.recipe_2.name,
                },
            ),
        ]

        self.creator = TupleJobCreator(
            inputs,
            self.out_experiment.name,
            self.out_recipe.name,
            tags=["test_tag"],
        )

    def test_get_existing_jobs(self):
        jobs = self.creator.get_existing_jobs()
        self.assertEqual(jobs.count(), 3)

    def test_get_input_queries(self):
        qs = self.creator.get_input_queries()

        self.assertIsInstance(qs, list)
        self.assertEqual(len(qs), 2)
        self.assertIsInstance(qs[0], list)
        self.assertEqual(len(qs[0]), 7)
        self.assertEqual(len(qs[1]), 5)

        node_id = qs[0][0]
        node = ChemNode.objects.get(id=node_id)
        self.assertEqual(node.parentjob.recipe, self.recipe_1)
        self.assertEqual(node.parentjob.experiment, self.inp_experiment)

        node_id = qs[1][0]
        node = ChemNode.objects.get(id=node_id)
        self.assertEqual(node.parentjob.recipe, self.recipe_2)
        self.assertEqual(node.parentjob.experiment, self.inp_experiment)

    def test_make_combinations(self):
        group1 = [1, 2]
        group2 = [3, 4, 5]

        combs = self.creator.make_combinations([group1, group2], exclude=[(1, 4)])

        self.assertEqual(len(combs), 5)

    def test_get_inputs_with_jobs(self):
        tuples = self.creator.get_inputs_with_jobs()
        self.assertEqual(len(tuples), 3)
        self.assertIsInstance(tuples[0], tuple)

    def test_get_inputs(self):
        nodes = self.creator.get_inputs()
        self.assertEqual(len(nodes), 32)

        for qs in nodes:
            self.assertEqual(qs.count(), 2)
            self.assertIsInstance(qs.first(), ChemNode)

    def test_create(self):
        jobs, inputs = self.creator.create()

        self.assertEqual(len(jobs), 32)

        job = jobs[0]
        self.assertEqual(job.inputs.count(), 2)
        self.assertEqual(job.tags.count(), 1)

        for j in jobs:
            j.delete()



from django.core.management.base import BaseCommand, CommandError

from mkite.orm.jobs.models import Project, Experiment


class Command(BaseCommand):
    help = "Adds a new experiment to the database"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "project",
            type=str,
            default=None,
            help="Name of the project where the experiment will be nested",
        )
        argparser.add_argument(
            "experiment",
            type=str,
            default=None,
            help="Name of the experiment to be created",
        )
        argparser.add_argument(
            "--dry_run",
            action="store_true",
            help="If set, does not store objects into the database",
        )
        return argparser

    def handle(self, project, experiment, *args, dry_run=False, **kwargs):
        prj = Project.objects.get(name=project)
        exp = Experiment.objects.filter(name=experiment, project=prj)

        if exp.exists():
            self.log(
                "error",
                f"An experiment named {experiment} already exists in project {project}.",
            )
            return

        if dry_run:
            self.log(
                "success",
                f"[DRY_RUN] Would have created experiment {experiment} under project {project}",
            )
        else:
            exp = Experiment.objects.create(name=experiment, project=prj)
            self.log("success", f"Added experiment {experiment} (id {exp.id})")



import json
from model_bakery import baker
from django.test import TestCase
from pkg_resources import resource_filename

from mkite.orm.base.models import Formula
from mkite.orm.jobs.models import Job
from mkite.orm.mols.models import Molecule, Conformer
from mkite.orm.mols.serializers import (
    MoleculeSerializer,
    ConformerSerializer,
)

from .test_models import MolCreator


TEST_CONFORMER = resource_filename("mkite_core.tests.files.models", "conformer.json")


class TestMolSerializer(TestCase):
    def setUp(self):
        self.creator = MolCreator()

    def test_serialize(self):
        mol = self.creator.create_molecule()
        serial = MoleculeSerializer(mol)
        data = serial.data

        expected = {
            "inchikey": "BSYNRYMUTXBXSQ-UHFFFAOYSA-N",
            "smiles": "CC(=O)OC1=CC=CC=C1C(=O)O",
            "attributes": {},
            "siteprops": {},
        }

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        job = baker.make(Job)
        data = {
            "parentjob": {"id": job.id},
            "smiles": "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
            "attributes": {},
            "siteprops": {},
        }
        serial = MoleculeSerializer(data=data)
        self.assertTrue(serial.is_valid())

        newmol = serial.save()
        self.assertEqual(newmol.smiles, "Cn1c(=O)c2c(ncn2C)n(C)c1=O")
        self.assertEqual(newmol.formula.name, "H10 C8 N4 O2 +0")
        self.assertEqual(newmol.parentjob.uuid, job.uuid)
        self.assertEqual(newmol.inchikey, "RYYVLZVUVIJVGH-UHFFFAOYSA-N")


class TestConformerSerializer(TestCase):
    def setUp(self):
        self.creator = MolCreator()

    def get_conf_dict(self):
        with open(TEST_CONFORMER, "r") as f:
            return json.load(f)

    def test_serialize(self):
        conf = self.creator.create_conformer()
        serial = ConformerSerializer(conf)
        data = serial.data

        expected = {
            "species": conf.species,
            "coords": conf.coords,
            "attributes": {},
            "siteprops": {},
        }

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        job = baker.make(Job)
        cdict = self.get_conf_dict()
        data = {
            "parentjob": {"id": job.id},
            "formula": cdict["formula"],
            "species": cdict["species"],
            "coords": cdict["coords"],
            "attributes": {},
            "siteprops": {},
        }
        serial = ConformerSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.coords, data["coords"])
        self.assertEqual(new.formula.name, data["formula"]["name"])
        self.assertEqual(new.parentjob.uuid, job.uuid)



from typing import List, Tuple

from django.db.models import QuerySet
from mkite.orm.base.models import ChemNode
from mkite.orm.jobs.models import Job

from .base import BaseJobCreator, JobCreationError


class SimpleJobCreator(BaseJobCreator):
    def get_inputs(self) -> "QuerySet[ChemNode]":
        if len(self.inputs) != 1:
            raise JobCreationError(
                f"Trying to use {self.__class__.__name__} with\
            more than one input specified. Please specify only one type\
            of input when using this class."
            )

        inp_kwargs = self.inputs[0]
        nodes = ChemNode.objects.filter(**inp_kwargs.filter)
        nodes = nodes.exclude(
            childjobs__experiment=self.out_experiment,
            childjobs__recipe=self.out_recipe,
            **inp_kwargs.exclude,
        )
        return nodes

    def create(self, dry_run: bool = False) -> Tuple[List[Job], QuerySet]:
        inputs = self.get_inputs()
        num_jobs = inputs.count()
        jobs = [self.job_template for n in range(num_jobs)]

        if dry_run:
            return jobs, inputs

        Job.objects.bulk_create(jobs, batch_size=self.batch_size)

        if self.tags:
            self.add_tags(jobs)

        JobParent = Job.inputs.through
        parents = [
            JobParent(chemnode_id=inp.id, job_id=job.id)
            for inp, job in zip(inputs, jobs)
        ]

        JobParent.objects.bulk_create(parents, batch_size=self.batch_size)

        return jobs, inputs



from .base import JobCreationError, InputQuery
from .simple import SimpleJobCreator
from .tuple import TupleJobCreator


JOB_CREATORS = {
    "simple": SimpleJobCreator,
    "tuple": TupleJobCreator,
}



import unittest as ut
from model_bakery import baker
from django.test import TestCase
from pkg_resources import resource_filename

from collections import namedtuple
from mkite_core.models import JobResults
from mkite.orm.jobs.models import Job, JobStatus, RunStats
from mkite.orm.structs.models import Crystal
from mkite.orm.calcs.models import EnergyForces

from mkite.workflow.parse import JobParser


RESULTS_FILE = resource_filename("mkite.tests.files.workflow", "jobresults.json")


class TestParser(TestCase):
    def setUp(self):
        self.results = JobResults.from_json(RESULTS_FILE)
        self.parser = JobParser(self.results)

    def test_create_stats(self):
        job = baker.make(Job, status="D")
        new = self.parser.create_stats(job)
        self.assertIsInstance(new, RunStats)

    def test_create_job(self):
        new = self.parser.create_job()
        self.assertIsInstance(new, Job)

    def test_create_chemnode(self):
        job = baker.make(Job, status="R")
        chemdict = self.results.nodes[0].chemnode

        new = self.parser.create_chemnode(chemdict, job)
        self.assertIsInstance(new, Crystal)

    def test_create_calcnode(self):
        chemnode = baker.make(Crystal)
        calcdict = self.results.nodes[0].calcnodes[0]

        new = self.parser.create_calcnode(calcdict, chemnode, chemnode.parentjob)
        self.assertIsInstance(new, EnergyForces)

    def test_create_nodes(self):
        job = baker.make(Job, status="R")
        nodes = self.parser.create_nodes(job)

        self.assertIsInstance(nodes, list)

        results = nodes[0]
        self.assertIsInstance(results.chemnode, Crystal)
        self.assertIsInstance(results.calcnodes, list)
        self.assertIsInstance(results.calcnodes[0], EnergyForces)

    def test_parse(self):
        out = self.parser.parse()

        self.assertIsInstance(out.job, Job)
        self.assertIsInstance(out.runstats, RunStats)
        self.assertIsInstance(out.nodes[0].chemnode, Crystal)
        self.assertIsInstance(out.nodes[0].calcnodes[0], EnergyForces)



import os
from typing import Iterable, List
from mkite_core.external import load_config

from ase import Atoms
from ase.io import read as ase_read
from mkite_core.models import CrystalInfo, ConformerInfo, MoleculeInfo, NodeResults
from .base import DbImporter, DbImporterError


class AseFileImporter(DbImporter):
    RECIPE_DICT = {
        "name": "dbimport.AseFileImporter",
        "method": "EXT",
    }
    PACKAGE_DICT = {"name": "mkite.dbimport"}

    def query(self, filename: os.PathLike, **kwargs):
        """Opens the file provided and returns its contents.
        For now, does not perform any filtering."""
        raw_data = self.read(filename)

        return raw_data

    def read(self, filename: os.PathLike):
        """Reads whatever ASE can read"""
        return ase_read(filename, index=":")

    def convert(self, parsed: List[dict]) -> List[NodeResults]:
        """Converts the files into Infos, which is then converted into a
        Node. 
        """
        return [self.convert_item(atoms) for atoms in parsed]

    def convert_item(self, atoms: Atoms) -> NodeResults:
        if atoms.pbc.any():
            info = self.get_crystal_info(atoms)
        else:
            info = self.get_conformer_info(atoms)

        return NodeResults(chemnode=info.as_dict())
    
    def get_crystal_info(self, atoms: Atoms) -> CrystalInfo:
        return CrystalInfo.from_ase(atoms)

    def get_conformer_info(self, atoms: Atoms) -> ConformerInfo:
        if "smiles" in atoms.info:
            mol = MoleculeInfo.from_smiles(atoms.info["smiles"])
        else:
            mol = None

        return ConformerInfo.from_ase(atoms, mol=mol)



import json
from typing import List
from django.db import transaction
from django.core.management.base import BaseCommand, CommandError

from mkite.workflow.create import InputQuery, SimpleJobCreator


class Command(BaseCommand):
    help = "Creates jobs according to the given input/output recipes"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "inp_experiment",
            type=str,
            help="Name of the experiment of the inputs",
        )
        argparser.add_argument(
            "inp_recipe",
            type=str,
            help="Name of the recipe of the ChemNodes to be considered as inputs",
        )
        argparser.add_argument(
            "out_experiment",
            type=str,
            help="Name of the experiment of the outputs",
        )
        argparser.add_argument(
            "out_recipe",
            type=str,
            default=None,
            help="Name of the recipe of the job to be executed",
        )
        argparser.add_argument(
            "-o",
            "--options",
            type=str,
            default=None,
            help="JSON string containing the options for the job to created",
        )
        argparser.add_argument(
            "-f",
            "--filter_kwargs",
            type=str,
            default=None,
            help="JSON string containing additional query options for selecting inputs",
        )
        argparser.add_argument(
            "-x",
            "--exclude_kwargs",
            type=str,
            default=None,
            help="JSON string containing additional query options for excluding inputs",
        )
        argparser.add_argument(
            "-t",
            "--tags",
            type=str,
            nargs="+",
            help="tags to be added to the jobs",
        )
        argparser.add_argument(
            "-b",
            "--batch_size",
            type=int,
            default=None,
            help="Size of the batch when bulk creating new database objects",
        )
        argparser.add_argument(
            "--dry_run",
            action="store_true",
            help="If set, does not store anything into the database",
        )
        return argparser

    @transaction.atomic
    def handle(
        self,
        inp_experiment,
        inp_recipe,
        out_experiment,
        out_recipe,
        *args,
        options=None,
        filter_kwargs=None,
        exclude_kwargs=None,
        tags=None,
        batch_size=None,
        dry_run=False,
        **kwargs,
    ):
        if options is not None:
            options = json.loads(options)

        inputs = self.get_input_query(
            inp_experiment=inp_experiment,
            inp_recipe=inp_recipe,
            filter_kwargs=filter_kwargs,
            exclude_kwargs=exclude_kwargs,
        )

        creator = SimpleJobCreator(
            inputs=inputs,
            out_experiment=out_experiment,
            out_recipe=out_recipe,
            options=options,
            tags=tags,
            batch_size=batch_size,
        )

        self.log("notice", "Constraints for job creation:")
        self.log("notice", f"Inputs: Experiment {inp_experiment}, Recipe {inp_recipe}")
        self.log("notice", f"Outputs: Experiment {out_experiment}, Recipe {out_recipe}")

        jobs, inputs = creator.create(dry_run=dry_run)

        msg = f"created {len(jobs)} new jobs."
        if dry_run:
            msg = "(DRY_RUN) would have" + msg

        self.log("success", msg)

    def get_input_query(
        self,
        inp_experiment: str,
        inp_recipe: str,
        filter_kwargs: str = None,
        exclude_kwargs: str = None,
    ) -> List[InputQuery]:
        filter_kwargs = {
            "parentjob__experiment__name": inp_experiment,
            "parentjob__recipe__name": inp_recipe,
            **self.decode_json_string(filter_kwargs),
        }
        exclude_kwargs = self.decode_json_string(exclude_kwargs)

        iq = InputQuery(filter=filter_kwargs, exclude=exclude_kwargs)

        return [iq]

    def decode_json_string(self, json_str: str):
        if json_str is None:
            return {}

        return json.loads(json_str)



import os
import unittest as ut
from mkite_core.external import load_config
from unittest.mock import MagicMock
from pkg_resources import resource_filename

from mkite_core.models import MoleculeInfo, NodeResults, JobResults
from mkite.dbimport.molfile import MolFileImporter


MOLFILE = resource_filename("mkite.tests.files.dbimport", "molfile.yaml")


class TestMolFileImporter(ut.TestCase):
    def setUp(self):
        self.dbimp = MolFileImporter(
            project="test_prj",
            experiment="test_exp",
        )

    def get_contents(self):
        return load_config(MOLFILE)

    def test_read(self):
        data = self.dbimp.read(MOLFILE)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 3)
        self.assertIsInstance(data[0], dict)

    def test_query(self):
        data = self.dbimp.query(filename=MOLFILE)
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 3)
        self.assertIsInstance(data[0], dict)

    def test_convert_item(self):
        data = self.get_contents()

        results = self.dbimp.convert_item(data[0])

        self.assertIsInstance(results, NodeResults)
        self.assertIsInstance(results.chemnode, dict)
        for key in ["smiles", "inchikey", "attributes"]:
            self.assertTrue(key in results.chemnode)

    def test_convert(self):
        data = self.get_contents()
        results = self.dbimp.convert(data)

        self.assertIsInstance(results, list)
        self.assertIsInstance(results[0], NodeResults)
        self.assertEqual(len(results), 3)



import unittest as ut

from mkite.orm.repr import _named_repr


class MockObject:
    @property
    def id(self):
        return 1

    @property
    def name(self):
        return "object"


class TestRepr(ut.TestCase):
    def setUp(self):
        self.obj = MockObject()

    def test_repr(self):
        out = _named_repr(self.obj)
        expected = "<MockObject: object (1)>"

        self.assertEqual(out, expected)



import unittest as ut
from io import StringIO
from django.test import TestCase
from django.core.management import call_command

from mkite.orm.jobs.management.commands.scanrecipes import Command


class TestCommand(TestCase):
    def call_command(self, *args, **kwargs):
        call_command(
            "scanrecipes",
            *args,
            stdout=StringIO(),
            stderr=StringIO(),
            **kwargs,
        )

    def test_command(self):
        self.call_command("--dry_run")



from mkite.orm.base.serializers import (
    FormulaSerializer,
    ChemNodeSerializer,
    CalcNodeSerializer,
)
from mkite.orm.calcs.serializers import (
    EnergyForcesSerializer,
    FeatureSerializer,
)
from mkite.orm.structs.serializers import CrystalSerializer
from mkite.orm.mols.serializers import (
    MoleculeSerializer,
    ConformerSerializer,
)
from mkite.orm.jobs.serializers import (
    ProjectSerializer,
    ExperimentSerializer,
    JobPackageSerializer,
    JobRecipeSerializer,
    RunStatsSerializer,
    JobSerializer,
)


SERIALIZERS = {
    "Formula": FormulaSerializer,
    "ChemNode": ChemNodeSerializer,
    "CalcNode": CalcNodeSerializer,
    "EnergyForces": EnergyForcesSerializer,
    "Feature": FeatureSerializer,
    "Molecule": MoleculeSerializer,
    "Conformer": ConformerSerializer,
    "Crystal": CrystalSerializer,
    "Project": ProjectSerializer,
    "Experiment": ExperimentSerializer,
    "JobPackage": JobPackageSerializer,
    "JobRecipe": JobRecipeSerializer,
    "RunStats": RunStatsSerializer,
    "Job": JobSerializer,
}


class DeserializeError(Exception):
    pass


def get_serializer(data: dict):
    modname = data.get("@module", None)
    clsname = data.get("@class", None)
    data = {k: v for k, v in data.items() if not k.startswith("@")}

    if clsname not in SERIALIZERS:
        raise DeserializeError(
            f"Could not find serializer for class {clsname}. Specify \
                a valid class module/name using the key `'@class'` in the \
                input data"
        )

    return SERIALIZERS[clsname](data=data)


class ModuleDeserializer:
    def __init__(self, data: dict):
        self.modname = data.get("@module", None)
        self.clsname = data.get("@class", None)
        self.data = {k: v for k, v in data.items() if not k.startswith("@")}

    def get_module(self):
        return __import__(self.modname, globals(), locals(), [self.clsname], 0)

    def save(self):
        if self.clsname is None or self.modname is None:
            raise DeserializeError(
                f"Could not find module {self.modname} for class {self.clsname}.\
                    specify valid module/class names using the keys `'@module'` \
                    /`'@class'` in the input data"
            )

        return self.deserialize()

    def deserialize(self):
        mod = self.get_module()
        if not hasattr(mod, self.clsname):
            raise DeserializeError(
                f"Class {self.clsname} not found in module {self.modname}"
            )

        cls_ = getattr(mod, self.clsname)

        if hasattr(cls_, "from_dict"):
            obj = cls_.from_dict(self.data)
            obj.save()
            return obj

        obj = cls_(**self.data)
        obj.save()
        return obj



import os
import json
from typing import List
from mkite_core.external import load_config
from django.core.management.base import BaseCommand, CommandError

from mkite import dbimport as dbimp
from mkite_core.models import JobResults
from mkite.workflow.parse import JobParser

from mkite.orm.jobs.models import Job


DB_IMPORTERS = {cls.__name__: cls for cls in dbimp.DbImporter.__subclasses__()}


class Command(BaseCommand):
    help = "Parses another database into the mkite database"

    def log(self, style, msg):
        style_fn = getattr(self.style, style.upper())
        return self.stdout.write(style_fn(msg))

    def add_arguments(self, argparser):
        argparser.add_argument(
            "importer",
            type=str,
            choices=DB_IMPORTERS.keys(),
            help="Name of the importer to be used",
        )
        argparser.add_argument(
            "-p",
            "--project",
            type=str,
            required=True,
            help="Name of the project where the data is going to be imported to",
        )
        argparser.add_argument(
            "-e",
            "--experiment",
            type=str,
            required=True,
            help="Name of the experiment where the data is going to be imported to",
        )
        argparser.add_argument(
            "-q",
            "--query",
            type=str,
            default=None,
            help="JSON string containing all arguments of the query to be \
                    performed",
        )
        argparser.add_argument(
            "-f",
            "--file",
            type=str,
            default=None,
            help="File containing the query to be processed. If present, \
                the command will ignore the JSON string passed with the -q \
                option. The file can be a yaml or json file",
        )
        return argparser

    def handle(self, importer, *args, **kwargs):
        self.project = kwargs["project"]
        self.experiment = kwargs["experiment"]

        importer_cls = DB_IMPORTERS[importer]
        self.importer = importer_cls.from_env(
            project=self.project,
            experiment=self.experiment,
        )
        self.queries = self.get_query_args(**kwargs)
        results = self.process_queries()
        self.save(results)

    def get_query_args(self, **kwargs):
        if kwargs.get("file", None) is not None:
            filename = kwargs["file"]
            if filename.endswith(".yaml") or filename.endswith(".json"):
                return load_config(filename)

            return {"filename": kwargs["file"]}

        if kwargs.get("query", None) is not None:
            return json.loads(kwargs["query"])

        raise CommandError(
            "No query specified. Please specify a query with the -q option, \
            or pass an input file with the -f option"
        )

    def process_queries(self) -> List[JobResults]:
        self.log("notice", f"Project: {self.project}")
        self.log("notice", f"Experiment: {self.experiment}")

        if isinstance(self.queries, dict):
            self.log(
                "notice",
                f"Importing single query using {self.importer.__class__.__name__}",
            )
            jobresults = self.importer.import_data(**self.queries)
            self.log("success", "Successfully queried the external database")
            return [jobresults]

        self.log(
            "notice",
            f"Importing multiple queries using {self.importer.__class__.__name__}",
        )

        results = []
        for qargs in self.queries:
            jobresults = self.importer.import_data(**qargs)
            results.append(jobresults)

        self.log("success", f"Imported data sets across {len(self.queries)}")
        return results

    def save(self, results: List[JobResults]):
        saved = 0
        for index, info in enumerate(results):
            try:
                self.save_jobresults(info)
                saved += 1
                self.log(
                    "success",
                    f"Imported query index {index}",
                )
            except Exception as e:
                self.log("error", f"Skipping import of query index {index}: {str(e)}")
                continue

        self.log(
            "success",
            f"Saved {saved} entries in the database",
        )

    def save_jobresults(self, info: JobResults):
        if not self.is_valid_parse(info):
            raise CommandError("Parsing the results of the query is not valid")

        return JobParser(info).parse()

    def is_valid_parse(self, info: JobResults) -> bool:
        """Verifies if it is valid to parse the JobResults given by info.
        This prevents duplicate jobs from being parsed into the system.
        """
        query = Job.objects.filter(
            experiment__name=self.experiment,
            experiment__project__name=self.project,
            options=info.job["options"],
        )

        return not query.exists()



from datetime import timedelta
from rest_framework import serializers

from taggit.serializers import TagListSerializerField, TaggitSerializer
from mkite.orm.serializers import BaseSerializer

from .models import Project, Experiment, Job, JobRecipe, JobPackage, RunStats


class ProjectSerializer(BaseSerializer):
    description = serializers.CharField(required=False)

    class Meta:
        model = Project
        fields = (
            "id",
            "uuid",
            "name",
            "description",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )


class ExperimentSerializer(BaseSerializer):
    project = ProjectSerializer(nested_field=True)
    description = serializers.CharField(required=False)

    class Meta:
        model = Experiment
        fields = (
            "id",
            "uuid",
            "name",
            "description",
            "project",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )


class JobPackageSerializer(BaseSerializer):
    class Meta:
        model = JobPackage
        fields = (
            "id",
            "uuid",
            "name",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )


class JobRecipeSerializer(BaseSerializer):
    package = JobPackageSerializer(nested_field=True)

    class Meta:
        model = JobRecipe
        fields = (
            "id",
            "uuid",
            "name",
            "method",
            "defaults",
            "package",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )


class RunStatsSerializer(BaseSerializer):
    class Meta:
        model = RunStats
        fields = (
            "id",
            "uuid",
            "host",
            "cluster",
            "duration",
            "ncores",
            "ngpus",
            "pkgversion",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )

    def convert_duration(self, duration) -> timedelta:
        if isinstance(duration, timedelta):
            return duration

        if type(duration) in [int, float]:
            duration = round(duration, 6)
            return timedelta(seconds=duration)

        return duration

    def create(self, validated_data: dict):
        validated_data["duration"] = self.convert_duration(validated_data["duration"])
        return super().create(validated_data)

    def update(self, instance, validated_data: dict):
        validated_data["duration"] = self.convert_duration(validated_data["duration"])
        return super().update(instance, validated_data)

    def to_representation(self, instance):
        data = super().to_representation(instance)
        data["duration"] = instance.duration.total_seconds()
        return data


class JobSerializer(TaggitSerializer, BaseSerializer):
    experiment = ExperimentSerializer(nested_field=True)
    recipe = JobRecipeSerializer(nested_field=True)
    runstats = RunStatsSerializer(nested_field=True, required=False)
    tags = TagListSerializerField(required=False)

    class Meta:
        model = Job
        fields = (
            "id",
            "uuid",
            "experiment",
            "recipe",
            "runstats",
            "status",
            "isroot",
            "options",
            "tags",
        )
        read_only_fields = (
            "ctime",
            "mtime",
        )



from datetime import timedelta
from django.db import models
from django.core.serializers.json import DjangoJSONEncoder

from mkite.orm.repr import _named_repr
from mkite.orm.base.models import DbEntry
from taggit.managers import TaggableManager


class Project(DbEntry):
    name = models.CharField(max_length=32, unique=True)
    description = models.CharField(max_length=256, null=True)

    def __repr__(self):
        return _named_repr(self)


class Experiment(DbEntry):
    name = models.CharField(max_length=32, unique=True)
    description = models.CharField(max_length=256, null=True)
    project = models.ForeignKey(
        Project,
        null=False,
        related_name="experiments",
        on_delete=models.PROTECT,
    )

    def __repr__(self):
        return _named_repr(self)


class JobStatus(models.TextChoices):
    # status for incomplete jobs
    READY = "Y"
    RUNNING = "R"
    STOPPED = "S"

    # status for complete jobs
    ERROR = "E"
    DONE = "D"


class Job(DbEntry):
    """Stores jobs in the database and its values"""

    experiment = models.ForeignKey(
        Experiment,
        null=False,
        db_index=True,
        related_name="jobs",
        on_delete=models.PROTECT,
    )

    inputs = models.ManyToManyField(
        "base.ChemNode",
        related_name="childjobs",
    )

    recipe = models.ForeignKey(
        "JobRecipe",
        null=False,
        db_index=True,
        related_name="jobs",
        on_delete=models.PROTECT,
    )

    runstats = models.OneToOneField(
        "RunStats", null=True, related_name="job", on_delete=models.CASCADE
    )

    status = models.CharField(
        max_length=1,
        null=False,
        choices=JobStatus.choices,
        default=JobStatus.READY,
    )

    options = models.JSONField(encoder=DjangoJSONEncoder, default=dict)

    isroot = models.BooleanField(default=False)

    tags = TaggableManager()

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.experiment.name}, {self.recipe.name}, {JobStatus(self.status).label} ({self.id})>"

    def as_dict(self):
        return {
            "id": self.id,
            "uuid": str(self.uuid),
            "ctime": self.ctime,
            "project": self.experiment.project.name,
            "experiment": self.experiment.name,
        }

    def as_info(self):
        from mkite_core.models import JobInfo

        return JobInfo.from_job(self)

    @property
    def parentjobs(self):
        return self.__class__.objects.filter(
            id__in=self.inputs.values_list("parentjob")
        )

    @property
    def childjobs(self):
        return self.__class__.objects.filter(
            id__in=self.chemnodes.values_list("childjobs")
        )


class RecipeMethods(models.TextChoices):
    """Enum class to provide different types of recipes.

    Attributes:
        DFT: Density Functional Theory
        FF: Force Field
        WF: Wavefunction
        ML: Machine Learning
        FT: Featurization
        EXT: External
        GEN: Generation
    """

    DFT = "DFT"
    FF = "FF"
    WF = "WF"
    ML = "ML"
    FT = "FT"
    EXT = "EXT"
    GEN = "GEN"


class JobRecipe(DbEntry):
    name = models.CharField(max_length=64, db_index=True)
    package = models.ForeignKey(
        "JobPackage", null=False, related_name="recipes", on_delete=models.PROTECT
    )
    method = models.CharField(
        max_length=3,
        null=False,
        choices=RecipeMethods.choices,
        default=RecipeMethods.DFT,
    )
    defaults = models.JSONField(encoder=DjangoJSONEncoder, default=dict)

    def as_dict(self):
        return {
            "id": self.id,
            "uuid": str(self.uuid),
            "name": self.name,
            "package": self.package.name,
            "method": str(self.method),
        }

    def __repr__(self):
        return _named_repr(self)


class JobPackage(DbEntry):
    name = models.CharField(max_length=64, db_index=True)

    def __repr__(self):
        return _named_repr(self)


class RunStats(DbEntry):
    host = models.CharField(max_length=64, db_index=True)
    cluster = models.CharField(max_length=64, db_index=True)
    duration = models.DurationField()
    ncores = models.PositiveIntegerField()
    ngpus = models.PositiveIntegerField(default=0)
    pkgversion = models.CharField(max_length=128, null=True)

    def as_dict(self):
        return {
            "host": self.host,
            "cluster": self.cluster,
            "duration": self.duration,
            "ncores": self.ncores,
            "ngpus": self.ngpus,
            "pkgversion": self.pkgversion,
        }

    def as_info(self):
        from mkite_core.models import RunStatsInfo

        return RunStatsInfo(**self.as_dict())

    @classmethod
    def from_info(cls, info: "RunStatsInfo") -> "RunStats":
        return cls(
            host=info.host,
            cluster=info.cluster,
            duration=info.duration,
            ncores=info.ncores,
            ngpus=info.ngpus,
            pkgversion=info.pkgversion,
        )

    @staticmethod
    def duration_from_seconds(seconds: float) -> timedelta:
        return timedelta(seconds=seconds)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.host} @ {self.cluster} ({self.id})>"



import os
import mp_api
import numpy as np
import unittest as ut
from unittest.mock import MagicMock
from monty.serialization import loadfn
from pkg_resources import resource_filename

from pymatgen.core import Structure
from mkite_core.models import CrystalInfo, EnergyForcesInfo, NodeResults, JobResults
from mkite.dbimport.mp import MPImporter, MP_KEY


RESPONSE_PATH = resource_filename("mkite.tests.files.dbimport", "mp.json")


class MockDoc(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class MockMPResponse(MagicMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.return_value = self.format_response()

    def format_response(self):
        response = loadfn(RESPONSE_PATH)

        return [MockDoc(**data) for data in response]


class MockMPImporter(MPImporter):
    query = MockMPResponse()


class TestMPImporter(ut.TestCase):
    def setUp(self):
        self.dbimp = MockMPImporter(
            project="test_prj",
            experiment="test_exp",
        )

    def get_query_args(self):
        return {
            "elements": ["Si"],
            "num_elements": (1, 1),
            "energy_above_hull": (0, 1e-6),
        }

    def get_response(self):
        qargs = self.get_query_args()
        return self.dbimp.query_structures(**qargs)

    def test_query_structures(self):
        response = self.get_response()
        self.assertTrue(self.dbimp.query.called)

    def test_convert_structure(self):
        response = self.get_response()
        doc = response[0]
        info = self.dbimp.convert_structure(doc)
        self.assertIsInstance(info, dict)

    def test_convert_energy_forces(self):
        response = self.get_response()
        doc = response[0]
        info = self.dbimp.convert_energy_forces(doc)
        self.assertIsInstance(info, dict)
        self.assertAlmostEqual(info["energy"], -5.42531402 * 2)
        self.assertTrue((np.array(info["forces"]) < 1e-10).all())

    def test_convert_doc(self):
        response = self.get_response()
        doc = response[0]
        info = self.dbimp.convert_doc(doc)
        self.assertIsInstance(info, NodeResults)

    def test_convert(self):
        response = self.get_response()
        info = self.dbimp.convert(response)
        self.assertIsInstance(info, list)

    def test_create_job(self):
        job = self.dbimp.create_job(
            project="test_prj",
            experiment="test_exp",
            isroot=True,
            options=self.get_query_args(),
        )
        expected = {
            "experiment": {"name": "test_exp", "project": {"name": "test_prj"}},
            "recipe": {
                "package": {"name": "mp_api.MPRester"},
                "name": "dbimport.MPImporter",
                "method": "EXT",
            },
            "isroot": True,
            "status": "D",
            "options": self.get_query_args(),
        }

        self.assertEqual(job, expected)

    def test_import_data(self):
        results = self.dbimp.import_data(
            project="test_prj",
            experiment="test_exp",
            isroot=True,
            **self.get_query_args()
        )
        self.assertIsInstance(results, JobResults)

    @ut.skip
    @ut.skipIf(MP_KEY not in os.environ, "MP_KEY not in environment")
    def test_connection(self):
        raise NotImplementedError("Test TO-DO")



import os
import numpy as np
from typing import Iterable, List

from mkite_core.models import CrystalInfo, EnergyForcesInfo, NodeResults
from .base import DbImporter, DbImporterError


MP_KEY = "MP_API_KEY"


class MPImporter(DbImporter):
    RECIPE_DICT = {
        "name": "dbimport.MPImporter",
        "method": "EXT",
    }
    PACKAGE_DICT = {"name": "mp_api.MPRester"}

    def __init__(self, project: str, experiment: str, api_key: str = None):
        super().__init__(project, experiment)
        self.key = api_key

    def get_rester(self):
        from mp_api import MPRester

        return MPRester(self.key)

    @classmethod
    def from_env(cls, project: str, experiment: str) -> "MPImporter":
        if MP_KEY not in os.environ:
            raise DbImporterError(f"{MP_KEY} is not defined in the environment")

        return cls(
            project=project,
            experiment=experiment,
            api_key=os.environ[MP_KEY],
        )

    def convert(self, docs: List[dict]) -> List[NodeResults]:
        return [self.convert_doc(doc) for doc in docs]

    def convert_doc(self, doc: dict) -> NodeResults:
        if not hasattr(doc, "structure"):
            raise DbImporterError(
                'Structure not found in MPImporter query.\
                Did you pass "structure" on the `fields` value of your \
                query?'
            )

        chemnode = self.convert_structure(doc)

        calcnodes = []
        if hasattr(doc, "energy_per_atom"):
            energy_forces = self.convert_energy_forces(doc)
            calcnodes.append(energy_forces)

        return NodeResults(chemnode=chemnode, calcnodes=calcnodes)

    def convert_structure(self, doc: dict) -> dict:
        structure = getattr(doc, "structure")
        info = CrystalInfo.from_pymatgen(structure)

        keys_to_save = ["material_id"]
        for k in keys_to_save:
            if hasattr(doc, k):
                info.attributes[k] = getattr(doc, k)

        return info.as_dict()

    def convert_energy_forces(self, doc: dict) -> dict:
        num_atoms = len(getattr(doc, "structure"))
        energy = getattr(doc, "energy_per_atom") * num_atoms
        forces = np.zeros((num_atoms, 3)).tolist()
        return EnergyForcesInfo(energy=energy, forces=forces).as_dict()

    def query(self, rester: str, query_function: str, **kwargs) -> Iterable[dict]:
        with self.get_rester() as mpr:
            engine = getattr(mpr, rester)
            fn = getattr(engine, query_function)
            docs = fn(**kwargs)

        return docs

    def query_structures(self, **kwargs) -> Iterable[dict]:
        fields = kwargs.get("fields", []) + [
            "material_id",
            "structure",
            "energy_per_atom",
        ]
        kwargs["fields"] = list(set(fields))

        return self.query("summary", "search", **kwargs)



import unittest as ut
from model_bakery import baker
from django.test import TestCase

from mkite.orm.base.models import ChemNode
from mkite.orm.base.serializers import ChemNodeSerializer
from mkite.orm.jobs.models import Job
from mkite.orm.jobs.serializers import JobSerializer
from mkite.orm.calcs.models import Feature, EnergyForces
from mkite.orm.calcs.serializers import EnergyForcesSerializer, FeatureSerializer


class TestFeatureSerializer(TestCase):
    @property
    def dict(self):
        return {"value": [1.0, 1.0, 0.0]}

    def test_serialize(self):
        f = baker.make(Feature, **self.dict)
        serial = FeatureSerializer(f)
        data = serial.data

        expected = {
            "@module": "mkite.orm.calcs.models",
            "@class": "Feature",
            **self.dict,
        }
        for k, v in expected.items():
            self.assertEqual(data[k], v)

        expected_keys = ["parentjob", "chemnode"]
        for key in expected_keys:
            self.assertTrue(key in data)

    def test_deserialize(self):
        chemnode = baker.make(ChemNode)
        data = {
            "@module": "mkite.orm.calcs.models",
            "@class": "Feature",
            "parentjob": {"id": chemnode.parentjob.id},
            "chemnode": {"id": chemnode.id},
            **self.dict,
        }
        serial = FeatureSerializer(data=data)

        self.assertTrue(serial.is_valid())

        new = serial.save()

        self.assertEqual(new.value, self.dict["value"])


class TestEnergyForcesSerializer(TestCase):
    def test_deserialize(self):
        chem = baker.make(ChemNode)
        data = {
            "@module": "mkite.orm.calcs.models",
            "@class": "EnergyForces",
            "parentjob": {"id": chem.parentjob.id},
            "chemnode": {"id": chem.id},
            "energy": -1,
            "forces": [[0, 0, 0]],
        }
        serial = EnergyForcesSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.chemnode, chem)
        self.assertEqual(new.parentjob, chem.parentjob)
        self.assertEqual(new.energy, data["energy"])
        self.assertEqual(new.forces, data["forces"])

    def test_serialize(self):
        node = baker.make(EnergyForces)
        jobdata = JobSerializer(node.parentjob).data
        chemdata = ChemNodeSerializer(node.chemnode).data
        expected = {
            "id": node.id,
            "uuid": str(node.uuid),
            "parentjob": jobdata,
            "chemnode": chemdata,
            "energy": node.energy,
            "forces": node.forces,
            "@module": "mkite.orm.calcs.models",
            "@class": "EnergyForces",
        }
        serial = EnergyForcesSerializer(node)

        self.assertEqual(serial.data, expected)



from django.db import transaction
from .models import Job


@transaction.atomic
def delete_tree(job: Job):
    """Cascades the deletion of the whole job tree under this job.
    All child jobs, along with all their nodes, will be deleted.
    """

    for j in job.childjobs:
        delete_tree(j)

    if job.runstats is not None:
        job.runstats.delete()

    job.calcnodes.all().delete()
    job.chemnodes.all().delete()
    job.delete()



from .base.models import ChemNode, CalcNode, Formula, Elements
from .jobs.models import (
    Project,
    Experiment,
    JobStatus,
    Job,
    JobRecipe,
    JobPackage,
    RunStats,
)
from .calcs.models import (
    EnergyForces,
    Feature,
)
from .mols.models import (
    Molecule,
    Conformer,
)
from .structs.models import (
    SpaceGroups,
    Crystal,
)



from datetime import timedelta
from django.test import TestCase
from mkite.orm.jobs.models import (
    Project,
    Experiment,
    JobStatus,
    Job,
    JobRecipe,
    JobPackage,
    RunStats,
)
from mkite_core.models import JobInfo, RunStatsInfo


class JobTestCreator:
    def create_project(self):
        prj, _ = Project.objects.get_or_create(
            name="test_prj", description="project for testing purposes"
        )
        return prj

    def create_experiment(self):
        prj = self.create_project()
        exp, _ = Experiment.objects.get_or_create(
            name="test_exp",
            description="first experiment",
            project=prj,
        )
        return exp

    def create_package(self):
        pkg, _ = JobPackage.objects.get_or_create(
            name="test_pkg",
        )
        return pkg

    def create_recipe(self):
        pkg = self.create_package()
        recipe, _ = JobRecipe.objects.get_or_create(
            name="test_recipe",
            package=pkg,
            method="DFT",
        )
        return recipe

    def create_runstats(self):
        stats, _ = RunStats.objects.get_or_create(
            host="test_host",
            cluster="test_cluster",
            duration=timedelta(seconds=3600),
            ncores=8,
            ngpus=0,
            pkgversion="1.0",
        )
        return stats

    def create_job(self):
        job, _ = Job.objects.get_or_create(
            experiment=self.create_experiment(),
            recipe=self.create_recipe(),
            runstats=self.create_runstats(),
            status=JobStatus.READY,
            isroot=True,
        )
        job.tags.add("tag1", "tag2")
        return job


class TestJobModels(TestCase):
    def setUp(self):
        self.creator = JobTestCreator()

    def test_project(self):
        prj = self.creator.create_project()
        newprj = Project.objects.get(name=prj.name)
        self.assertEqual(newprj, prj)

    def test_experiment(self):
        exp = self.creator.create_experiment()
        newexp = Experiment.objects.get(name=exp.name)
        self.assertEqual(exp, newexp)

    def test_package(self):
        pkg = self.creator.create_package()
        newpkg = JobPackage.objects.get(name=pkg.name)
        self.assertEqual(pkg, newpkg)

    def test_recipe(self):
        recipe = self.creator.create_recipe()
        newrecipe = JobRecipe.objects.get(name=recipe.name)
        self.assertEqual(recipe, newrecipe)

    def test_recipe_dict(self):
        recipe = self.creator.create_recipe()

        rdict = recipe.as_dict()
        self.assertTrue("uuid" in rdict)
        self.assertEqual(rdict["name"], recipe.name)
        self.assertEqual(rdict["package"], "test_pkg")
        self.assertEqual(rdict["method"], "DFT")

    def test_job(self):
        job = self.creator.create_job()
        newjob = Job.objects.get(
            status=JobStatus.READY, isroot=True, recipe__name="test_recipe"
        )
        self.assertEqual(newjob, job)

    def test_job_info(self):
        job = self.creator.create_job()
        info = job.as_info()
        self.assertIsInstance(info, JobInfo)

    def test_job_dict(self):
        job = self.creator.create_job()
        data = job.as_dict()
        expected = {
            "id": job.id,
            "uuid": str(job.uuid),
            "ctime": job.ctime,
            "project": job.experiment.project.name,
            "experiment": job.experiment.name,
        }
        self.assertEqual(data, expected)


class TestRunStats(TestCase):
    def setUp(self):
        self.creator = JobTestCreator()
        self.stats = self.creator.create_runstats()

    def test_create(self):
        self.assertFalse(self.stats.id is None)
        self.assertEqual(self.stats.host, "test_host")
        self.assertEqual(self.stats.cluster, "test_cluster")

    def test_dict(self):
        expected = {
            "host": "test_host",
            "cluster": "test_cluster",
            "duration": timedelta(seconds=3600),
            "ncores": 8,
            "ngpus": 0,
            "pkgversion": "1.0",
        }

        self.assertEqual(self.stats.as_dict(), expected)

    def test_info(self):
        info = self.stats.as_info()
        self.assertIsInstance(info, RunStatsInfo)

        new = RunStats.from_info(info)

    def test_duration(self):
        t = 1800
        duration = RunStats.duration_from_seconds(t)

        self.assertIsInstance(duration, timedelta)
        self.assertEqual(duration, timedelta(seconds=t))



from django.db import transaction
from rest_framework import serializers

from taggit.serializers import TagListSerializerField, TaggitSerializer
from mkite.orm.serializers import BaseSerializer
from mkite.orm.base.models import Formula
from mkite.orm.base.serializers import FormulaSerializer, ChemNodeSerializer
from mkite_core.models import FormulaInfo, CrystalInfo, SpaceGroupInfo

from .models import Crystal, SpaceGroups


class CrystalSerializer(TaggitSerializer, ChemNodeSerializer):
    spacegroup = serializers.ChoiceField(SpaceGroups.choices, required=False)
    formula = FormulaSerializer(nested_field=True, required=False)
    tags = TagListSerializerField(required=False)

    class Meta:
        model = Crystal
        fields = "__all__"

    @transaction.atomic
    def create(self, validated_data):
        if "formula" not in validated_data:
            info = FormulaInfo.from_list(validated_data["species"])
            validated_data["formula"] = info.as_dict()

        if "spacegroup" not in validated_data:
            info = CrystalInfo.from_dict(validated_data)
            spgrp = SpaceGroupInfo.from_info(info)
            validated_data["spacegroup"] = spgrp.number

        return super().create(validated_data)



import os
import json
import shutil
import unittest as ut
from io import StringIO
from model_bakery import baker
from django.test import TestCase
from django.db import models
from django.core.management import call_command
from pkg_resources import resource_filename

from mkite.orm.jobs.models import Job, JobStatus, JobRecipe, Experiment, Project
from mkite_core.tests.tempdirs import run_in_tempdir
from mkite.workflow.management.commands.submit import Command, CommandError

from mkite_core.tests.tempdirs import run_in_tempdir
from mkite_engines import LocalProducer


ENGINE = resource_filename("mkite.tests.files", "engine.yaml")


class TestCommand(TestCase):
    def setUp(self):
        self.cmd = Command(stdout=StringIO(), stderr=StringIO())
        self.cmd.recipe = "test_recipe"
        self.cmd.project = "test_project"
        self.cmd.experiment = "test_experiment"

    def call_command(self, *args, **kwargs):
        call_command(
            "submit",
            *args,
            stdout=StringIO(),
            stderr=StringIO(),
            **kwargs,
        )

    def test_get_recipe_args(self):
        recipe = baker.make(JobRecipe, name="test_recipe")
        args = self.cmd.get_recipe_args()

        self.assertEqual(args, {"recipe": recipe.id})
        recipe.delete()

        self.cmd.recipe = None
        args = self.cmd.get_recipe_args()
        self.assertEqual(args, {})

    def test_get_experiment_args(self):
        prj = baker.make(Project, name="test_project")
        exp = baker.make(Experiment, name="test_experiment", project=prj)

        args = self.cmd.get_experiment_args()
        self.assertEqual(args, {"experiment": exp.id})

        self.cmd.experiment = None
        args = self.cmd.get_experiment_args()
        self.assertEqual(args, {"experiment__project": prj.id})

        self.cmd.project = None
        self.cmd.experiment = "test_experiment"
        args = self.cmd.get_experiment_args()
        self.assertEqual(args, {"experiment": exp.id})

        self.cmd.project = None
        self.cmd.experiment = None
        args = self.cmd.get_experiment_args()
        self.assertEqual(args, {})

        self.cmd.project = "nonexisting"
        with self.assertRaises(CommandError):
            self.cmd.get_experiment_args()

        exp.delete()
        prj.delete()

    def test_get_jobs(self):
        prj = baker.make(Project, name="test_project")
        exp = baker.make(Experiment, name="test_experiment", project=prj)
        recipe = baker.make(JobRecipe, name="test_recipe")

        njobs = 5
        jobs = baker.make(
            Job, njobs, recipe=recipe, experiment=exp, status=JobStatus.READY
        )

        self.cmd.recipe = recipe.name
        self.cmd.experiment = exp.name
        self.cmd.project = prj.name

        query = self.cmd.get_jobs()

        self.assertEqual(query.count(), njobs)

    @run_in_tempdir
    def test_command(self):
        prj = baker.make(Project)
        exp = baker.make(Experiment, project=prj)
        recipe = baker.make(JobRecipe)

        njobs = 5
        jobs = baker.make(
            Job, njobs, recipe=recipe, experiment=exp, status=JobStatus.READY
        )

        self.call_command(
            ENGINE,
            "--project",
            prj.name,
            "--experiment",
            exp.name,
            "--recipe",
            recipe.name,
        )



from typing import Union
from collections import namedtuple
from django.db import transaction

from mkite_core.models import JobResults
from mkite.orm.deserializers import get_serializer, DeserializeError
from mkite.orm.jobs.serializers import JobSerializer, RunStatsSerializer


ParserOutput = namedtuple("ParserOutput", "job runstats nodes")
NodesOutput = namedtuple("NodesOutput", "chemnode calcnodes")


class JobParser:
    """Class that parses the jobs correctly run in different systems.
    Given a certain engine, gets all jobs that have been correctly run
    and not yet parsed.
    """

    def __init__(self, results: JobResults):
        self.results = results

    @transaction.atomic
    def parse(self) -> ParserOutput:
        job = self.create_job()
        runstats = self.create_stats(job)

        if runstats is not None:
            job.runstats = runstats
            job.save()

        nodes = self.create_nodes(job)

        return ParserOutput(job=job, runstats=runstats, nodes=nodes)

    def create_job(self) -> "Job":
        data = self.results.job

        if "status" not in data:
            data["status"] = "D"

        serial = JobSerializer(data=data)
        if not serial.is_valid():
            raise DeserializeError(f"Error deserializing Job. Errors: {serial.errors}")

        return serial.save()

    def create_stats(self, job: "Job") -> "RunStats":
        if not self.results.runstats:
            return None

        data = self.results.runstats.as_dict()

        if type(data.get("duration")) == float:
            data["duration"] = round(data["duration"], 6)

        serial = RunStatsSerializer(data=data)

        if not serial.is_valid():
            raise DeserializeError(
                f"Error deserializing RunStats. Errors: {serial.errors}"
            )

        return serial.save()

    def create_nodes(self, job: "Job") -> NodesOutput:
        nodes = []
        for node_results in self.results.nodes:
            chnode = self.create_chemnode(node_results.chemnode, job=job)
            calcs = [
                self.create_calcnode(calcdict, job=job, chemnode=chnode)
                for calcdict in node_results.calcnodes
            ]
            nodes += [NodesOutput(chemnode=chnode, calcnodes=calcs)]

        return nodes

    def create_chemnode(self, chemdict: dict, job: "Job") -> "ChemBase":
        chemdict["parentjob"] = {"id": job.id}
        return self.deserialize_node(chemdict)

    def create_calcnode(
        self, calcdict: dict, chemnode: "ChemBase", job: "Job"
    ) -> "CalcBase":
        calcdict["parentjob"] = {"id": job.id}
        calcdict["chemnode"] = {"id": chemnode.id}
        return self.deserialize_node(calcdict)

    def deserialize_node(self, nodedict: dict) -> Union["ChemBase", "CalcBase"]:
        serial = get_serializer(nodedict)
        if not serial.is_valid():
            raise DeserializeError(f"Error deserializing Job. Errors: {serial.errors}")
        return serial.save()



from django.db import models
from django.contrib.postgres.fields import ArrayField

from mkite.orm.base.models import CalcNode


class EnergyForces(CalcNode):
    energy = models.FloatField(null=True)
    forces = ArrayField(ArrayField(models.FloatField(), size=3), null=True)


class Feature(CalcNode):
    value = ArrayField(models.FloatField())



from django.db import transaction
from rest_framework import serializers

from mkite.orm.base.serializers import CalcNodeSerializer

from .models import EnergyForces, Feature


class EnergyForcesSerializer(CalcNodeSerializer):
    class Meta:
        model = EnergyForces
        fields = (
            "id",
            "uuid",
            "energy",
            "forces",
            "parentjob",
            "chemnode",
        )


class FeatureSerializer(CalcNodeSerializer):
    class Meta:
        model = Feature
        fields = (
            "id",
            "uuid",
            "value",
            "parentjob",
            "chemnode",
        )



from model_bakery import baker
from django.test import TestCase

from mkite.orm.base.models import Formula
from mkite.orm.jobs.models import Job
from mkite.orm.structs.serializers import CrystalSerializer

from .test_models import CrystalCreator


class TestCrystalSerializer(TestCase):
    def setUp(self):
        self.creator = CrystalCreator()

    def test_serialize(self):
        crystal = self.creator.create_crystal()
        serial = CrystalSerializer(crystal)
        data = serial.data

        expected = {
            "spacegroup": 227,
            "species": ["Si", "Si"],
            "coords": [[0.0, 0.0, 0.0], [1.365, 1.365, 1.365]],
            "lattice": [[0.0, 2.73, 2.73], [2.73, 0.0, 2.73], [2.73, 2.73, 0.0]],
            "attributes": {},
            "siteprops": {},
        }

        for k, v in expected.items():
            self.assertEqual(data[k], v)

    def test_deserialize(self):
        job = baker.make(Job)
        data = {
            "parentjob": {"id": job.id},
            "spacegroup": 227,
            "species": ["Si", "Si"],
            "coords": [[0.0, 0.0, 0.0], [1.365, 1.365, 1.365]],
            "lattice": [[0.0, 2.73, 2.73], [2.73, 0.0, 2.73], [2.73, 2.73, 0.0]],
            "attributes": {},
            "siteprops": {},
        }
        serial = CrystalSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.spacegroup, 227)
        self.assertEqual(new.formula.name, "Si2 +0")
        self.assertEqual(new.parentjob.uuid, job.uuid)

    def test_deserialize_without_spgrp(self):
        job = baker.make(Job)
        data = {
            "parentjob": {"id": job.id},
            "species": ["Si", "Si"],
            "coords": [[0.0, 0.0, 0.0], [1.365, 1.365, 1.365]],
            "lattice": [[0.0, 2.73, 2.73], [2.73, 0.0, 2.73], [2.73, 2.73, 0.0]],
            "attributes": {},
            "siteprops": {},
        }
        serial = CrystalSerializer(data=data)
        self.assertTrue(serial.is_valid())

        new = serial.save()
        self.assertEqual(new.spacegroup, 227)



from typing import Dict, Iterable
from itertools import chain
from django.db import models, transaction
from rest_framework import serializers
from rest_framework.fields import empty


class BaseSerializer(serializers.ModelSerializer):
    """Class that augments the functionalities of DRF to automate
    the creation/update of instances if unique fields are
    specified in the input data. Instead of attempting the creation
    of a new instance every time a serializer is used (POST behavior),
    this class simplifies the representation of the input data, which
    enables `id`, `uuid`, and `name` fields to be used as identifiers
    to check the existence of a new instance.
    """

    id = serializers.IntegerField(required=False)
    uuid = serializers.UUIDField(required=False)

    def __init__(self, instance=None, data=empty, nested_field=False, **kwargs):
        if instance is None and data is not empty:
            instance, data = self.get_instance_from_data(data)

        if instance is not None:
            kwargs["partial"] = kwargs.get("partial", True)

        super().__init__(instance, data, **kwargs)

        self.nested_field = nested_field
        if nested_field:
            self._setup_serializer_as_field()

    def _setup_serializer_as_field(self):
        """Useful to bypass validation when nested serializers are used
        in mkite. Ensures then that the rest of the validation is
        performed by the user at the creation/update stage.
        """
        for field_name, field in self.fields.items():
            field.required = False
            field.read_only = False
            if field_name in self.id_fields:
                field.validators = []

    @property
    def id_fields(self):
        default = ["id", "uuid", "name", "inchikey", "smiles"]
        return getattr(self.Meta, "id_fields", default)

    def get_instance_from_data(self, data: dict):
        model = self.Meta.model

        id_data = {k: v for k, v in data.items() if k in self.id_fields}
        if not id_data:
            return None, data

        instances = model.objects.filter(**id_data)
        if not instances.exists():
            return None, data

        if instances.count() > 1:
            raise serializers.ValidationError(
                "There is more than one \
                instance for the data provided."
            )

        instance = instances.first()

        # prevent the serializer from passing unique-like data ahead
        new_data = {k: v for k, v in data.items() if k not in self.id_fields}

        return instance, new_data

    def to_representation(self, instance):
        data = super().to_representation(instance)
        data["@module"] = instance._meta.model.__module__
        data["@class"] = instance._meta.model.__name__
        return data

    def get_nested_fields(self) -> Dict[str, serializers.Field]:
        return {
            name: field
            for name, field in self.fields.items()
            if getattr(field, "nested_field", False)
        }

    def get_nested_objects(self, validated_data: dict) -> Dict[str, object]:
        """Using the given validated data, retrieves all information
        that had been passed to the serializer to create/get the model
        corresponding to that field. As this base serializer has plenty
        of nested fields, this model helps creating/updating new objects
        based on these nested fields.
        """
        nested_objs = {}

        nested_fields = self.get_nested_fields()

        for field_name, field in nested_fields.items():
            if field_name not in validated_data:
                continue

            field_data = validated_data.pop(field_name)
            obj = self.deserialize_nested(field, field_data)
            nested_objs[field_name] = obj

        return nested_objs

    def update_m2m(self, instance, field_name: str, field_value: Iterable):
        field = getattr(instance, field_name)
        if not isinstance(field_value, Iterable):
            field_value = [field_value]

        field.add(*field_value)

    @transaction.atomic
    def create(self, validated_data: dict):
        nested_objs = self.get_nested_objects(validated_data)
        create_dict = {**validated_data, **nested_objs}

        model = self.Meta.model
        instance = model.objects.create(**create_dict)

        return instance

    @transaction.atomic
    def update(self, instance, validated_data: dict):
        nested_objs = self.get_nested_objects(validated_data)
        update_dict = {**validated_data, **nested_objs}

        for k, v in update_dict.items():
            setattr(instance, k, v)

        instance.save()

        return instance

    def deserialize_nested(self, field, field_data: dict):
        serializer = field.__class__(data=field_data)
        if not serializer.is_valid():
            raise serializers.ValidationError(
                f"Data for {field.field_name} is not valid"
            )

        return serializer.save()
